from neo import io
from collections import OrderedDict
import os
import json
import hashlib
import calendar
import re
import six
import logging
import quantities as pq


class manageFiles:
    @classmethod
    def md5(cls, filename):
        '''
        Generate hash md5 code for the filename passed as parameter
        '''
        hash_md5 = hashlib.md5()

        with open(filename, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()


class manageMetadata:
    """
    Class to read and manage information from metadata
    """

    # pattern for string substitution
    pattern = re.compile(r"[ \\.,\/]")

    @classmethod
    def get_cell_info(cls, filename_meta):
        """
        Extract cell info from the metadata file
        """

        try:
            # if a .json metadata file exists
            with open(filename_meta) as f:
                data = json.load(f)

                if "animal_species" not in data or \
                        data["animal_species"] is None or \
                        data["animal_species"].lower() == "unknown":
                    c_species = "unknown_species"
                else:
                    c_species = data["animal_species"]
                    c_species = cls.pattern.sub("-", c_species)
                    # c_species = c_species.replace(" ", "-")

                if "brain_structure" not in data or \
                        data["brain_structure"] is None or \
                        data["brain_structure"].lower() == "unknown":
                    c_area = "unknown_area"
                else:
                    c_area = data["brain_structure"]
                    c_area = cls.pattern.sub("-", c_area)

                if "cell_soma_location" not in data or \
                        data["cell_soma_location"] is None or \
                        data["cell_soma_location"].lower() == "unknown":
                    c_region = "unknown_region"
                else:
                    c_region = data["cell_soma_location"]
                    c_region = cls.pattern.sub("-", c_region)

                if "cell_type" not in data or data["cell_type"] is None or \
                        data["cell_type"].lower() == "unknown":
                    c_type = "unknown_type"
                else:
                    c_type = data["cell_type"]
                    c_type = cls.pattern.sub("-", c_type)

                if "etype" not in data or data["etype"] is None or \
                        data["etype"] == "unknown":
                    c_etype = "unknown_etype"
                else:
                    c_etype = data["etype"]
                    c_etype = cls.pattern.sub("-", c_etype)

                if "contributors_affiliations" not in data or \
                        data["contributors_affiliations"] is None or \
                        data["contributors_affiliations"].lower() == "unknown":
                    c_contrib = "unknown_contrib"
                else:
                    c_contrib = data["contributors_affiliations"]
                    c_contrib = cls.pattern.sub("-", c_contrib)

                if "cell_id" not in data or data["cell_id"] is None or \
                        data["cell_id"] == "unknown":
                    head, c_name = \
                        os.path.split(os.path.split(filename_meta)[0])
                else:
                    c_name = data["cell_id"]
                    c_name = cls.pattern.sub("-", c_name)

                if "filename" not in data or data["filename"] is None or \
                        data["filename"] == "unknown":
                    base = os.path.basename(os.path.normpath(filename_meta))
                    c_sample = os.path.splitext(base)[0]

                else:
                    c_sample = data["filename"]
                    c_sample = os.path.splitext(c_sample)[0]
                    c_sample = cls.pattern.sub("-", c_sample)

        except Exception as e:
            # if there is no metadata
            c_area = "unknown_area"
            c_species = "unknown_species"
            c_region = "unknown_region"
            c_type = "unknown_type"
            c_etype = "unknown_etype"
            c_contrib = "unknown_contributors"

            abf_meta_fn = cls.get_abf_filename(filename_meta)
            base = os.path.basename(os.path.normpath(abf_meta_fn))
            head, c_name = os.path.split(os.path.split(abf_meta_fn)[0])
            c_sample = os.path.splitext(base)[0]

        return (c_species, c_area, c_region, c_type, c_etype,
                c_contrib, c_name, c_sample)

    @classmethod
    def get_metadata(cls, metadata_file):
        """
        Read metadata file into a json dictionary
        """
        logging.info("Reading metadata " + os.path.basename(metadata_file))

        crr_file_dir = os.path.dirname(__file__)
        default_metadata_file = os.path.join(
            crr_file_dir, 'configs', 'metadata_template.json')
        with open(default_metadata_file, 'r') as mf:
            metadata = json.load(mf)
        mf.close()

        if os.path.exists(metadata_file):
            try:
                with open(metadata_file) as f:
                    metadata = json.load(f)
            except Exception as e:
                pass

        return metadata

    @classmethod
    def get_metadata_filename(cls, filename):
        """
        Build metadata filename, based on data filename
        """
        filepath, name = os.path.split(filename)
        name_no_ext, extension = os.path.splitext(name)
        metadata_file = os.path.join(filepath, name_no_ext + '_metadata.json')

        return metadata_file

    @classmethod
    def get_abf_filename(cls, filename_meta):
        """
        Build data filename based on metadata filename
        """

        filepath, name = os.path.split(filename_meta)
        name_no_ext, extension = os.path.splitext(name)
        newname = name_no_ext.replace('_metadata', '')
        abf_file = os.path.join(filepath, newname + '.abf')

        return abf_file

    @classmethod
    def extract_authorized_collab(cls, metadata_file):
        """
        Extract authorized collab from metadata file
        """

        with open(metadata_file) as meta:
            all_meta = json.load(meta)

        return all_meta['authorized_collabs']

    @classmethod
    def generate_json_data(cls, obj, outfilename, json_dir):
        """
        Dump json data to file
        """
        if not os.path.exists(json_dir):
            os.makedirs(json_dir)

        outfilepath = os.path.join(json_dir, outfilename)

        if not os.path.isfile(outfilepath):
            with open(outfilepath, 'w') as f:
                json.dump(obj, f)
        return

    @classmethod
    def generate_authorization_json(
            cls, final_json_name_list, metadatalist, final_dir):

        files_authorization = {}
        crr_file_auth_collab = extract_authorized_collab(metadatalist)
        files_authorization[final_json_name_list] = crr_file_auth_collab

        file_auth_fullpath = os.path.join(
            final_dir, "files_authorization.json")
        with open(file_auth_fullpath, 'a') as fa:
            json.dump(files_authorization, fa)

        return True

    @classmethod
    def stim_feats_from_meta(cls, crr_dict, num_segments):

        # read stimulus information
        try:
            if 'filename' not in crr_dict or not crr_dict['filename'] or \
                    crr_dict["filename"] == "unknown":
                raise Exception("'filename' key absent in metadata")

            if 'stimulus_type' not in crr_dict or \
                    crr_dict['stimulus_type'] != "step" or \
                    crr_dict["stimulus_type"] == "unknown":
                raise Exception("'stimulus_type' key absent in metadata \
                        for file:" + crr_dict['filename'])
            else:
                ty = str(crr_dict['stimulus_type'])
                logging.info("extracted stimulus type")

            if 'stimulus_time_unit' not in crr_dict or not \
                    crr_dict['stimulus_time_unit'] or \
                    crr_dict["stimulus_time_unit"] == "unknown":
                raise Exception("'stimulus_time_unit' key absent in metadata \
                    for file:" + crr_dict['filename'])
            else:
                tu = crr_dict['stimulus_time_unit']
                logging.info("extracted stimulus unit")

            if 'stimulus_start' in crr_dict \
                    and crr_dict['stimulus_start'] and \
                    crr_dict["stimulus_start"] != "unknown" and \
                    'stimulus_end' in crr_dict and \
                    crr_dict['stimulus_end'] and \
                    crr_dict['stimulus_end'] != 'uknown':
                st = crr_dict['stimulus_start'][0]
                en = crr_dict['stimulus_end'][0]
                logging.info("extracted stimulus start and stimulus \
                        end")
            elif 'tamp' in crr_dict and crr_dict['tamp']:
                st = crr_dict['tamp'][0]
                en = crr_dict['tamp'][1]
                logging.info("extracted stimulus start")
            elif 'ton' in crr_dict and 'toff' in crr_dict:
                st = crr_dict['ton']
                en = crr_dict['toff']
                logging.info("extracted stimulus start")
            else:
                raise Exception("'stimulus_start' and/or 'stimulus_end' key \
                        absent in metadata, for file:" + crr_dict['filename'])

            if 'stimulus_unit' in crr_dict and crr_dict['stimulus_unit'] and \
                    crr_dict["stimulus_unit"] != "unknown":
                u = str(crr_dict['stimulus_unit'])
                logging.info("extracted stimulus unit")
            elif 'i_unit' in crr_dict and crr_dict['i_unit'] and \
                    crr_dict['i_unit'] != 'unknown':
                u = str(crr_dict['i_unit'])
                logging.info("extracted stimulus unit")
            else:
                raise Exception("'stimulus_unit' key absent in metadata, \
                        for file:" + crr_dict['filename'])

            if 'stimulus_first_amplitude' not in crr_dict or not \
                    crr_dict['stimulus_first_amplitude'] or \
                    crr_dict["stimulus_first_amplitude"] == "unknown":
                raise Exception("'stimulus_first_amplitude' key absent in \
                        metadata, for file:" + crr_dict['filename'])
            else:
                fa = float(
                    format(
                        crr_dict['stimulus_first_amplitude'][0],
                        '.3f'))
                logging.info("extracted stimulus first amplitude")

            if 'stimulus_increment' not in crr_dict \
                    or crr_dict["stimulus_increment"] == "unknown":
                raise Exception("'stimulus_increment' key absent in metadata, \
                        for file:" + crr_dict['filename'])
            elif not crr_dict['stimulus_increment']:
                inc = float(format(0, '.3f'))
            else:
                inc = float(format(crr_dict['stimulus_increment'][0], '.3f'))
            logging.info("extracted stimulus increment")

            if 'sampling_rate_unit' not in crr_dict or not \
                    crr_dict['sampling_rate_unit'] \
                    or crr_dict["sampling_rate_unit"] == "unknown":
                ru = 'Hz'
                logging.info("extracted sampling rate unit")
            else:
                ru = crr_dict['sampling_rate_unit'][0]
                logging.info("extracted sampling rate unit")

            if 'sampling_rate' not in crr_dict or not \
                    crr_dict['sampling_rate'] or \
                    crr_dict["sampling_rate"] == "unknown":
                raise Exception("sampling_rate key absent in metadata, for \
                        file:" + crr_dict['filename'])
            else:
                r = crr_dict['sampling_rate'][0]
                logging.info("extracted sampling rate")

            if tu == 's':
                st = st * 1e3
                en = en * 1e3
        except Exception as e:
            logger.error(
                'Error in reading keys for stimulus extraction in file: ' +
                crr_dict['filename']
            )
            logger.error('ERROR ' + str(e))

        all_stim_feats = {
            "ty": [],
            "st": [],
            "en": [],
            "crr_val": [],
            "u": [],
            "ru": [],
            "r": []
        }

        if not fa:
            fa = 0

        # for every segment in the data file, compute stimulus amplitude
        for i in range(num_segments):
            crr_val = float(format(fa + inc * float(format(i, '.3f')), '.3f'))
            all_stim_feats["ty"].append(ty)
            all_stim_feats["st"].append(st)
            all_stim_feats["en"].append(en)
            all_stim_feats["crr_val"].append(crr_val)
            all_stim_feats["u"].append(u)
            all_stim_feats["r"].append(r)
            all_stim_feats["ru"].append(ru)

        return (1, all_stim_feats)

    @classmethod
    def generate_citation_json(cls, jsonfilenamelist, metadatalist, final_dir):
        """
        Generate a file with 'How to cite' information
        """
        citation_json = {}
        crrjsonfile = jsonfilenamelist
        with open(metadatalist) as cf:
            cmd = json.load(cf)
            caff = cmd['contributors_affiliations']
            cref = cmd['reference']
            if crrjsonfile not in citation_json:
                citation_json[crrjsonfile] = {cref: [caff]}
            else:
                pass

        with open(os.path.join(final_dir, "citation_list.json"), 'a') as f:
            json.dump(citation_json, f)

    @classmethod
    def get_contributors(cls, crr_dict):
        """
        Extract contributors name and affiliation from dictionary
        """

        # fill contributors field
        if "contributors_affiliations" in crr_dict:
            crr_contr = crr_dict["contributors_affiliations"]
            contributors = {
                'name': crr_contr, "message": "Data contributors: " +
                crr_contr}
        else:
            contributors = {}

        return contributors

    @classmethod
    def get_ljp(cls, crr_dict, volt_unit, ljpflag):
        """
        Extract the liquid junction potential (i.e. ljp)
        """

        ljp = {}

        # fill ljp (i.e. liquid junction potential) field
        if "liquid_junction_potential" in crr_dict and \
                "liquid_junction_potential_unit" in crr_dict:
            crr_ljp = crr_dict["liquid_junction_potential"]
            crr_ljp_u = crr_dict["liquid_junction_potential_unit"]

            if crr_ljp and crr_ljp[0] != 0 and crr_ljp_u:
                crr_ljp = crr_ljp * conversion_factor(volt_unit, crr_ljp_u)
                ljp = {
                    "ljp": {
                        "value": crr_ljp, "unit": crr_ljp_u,
                        "message": "Liquid junction potential value: " +
                        str(crr_ljp[0]), "computedflag": ljpflag}}

                return ljp

    @classmethod
    def get_holding_current(cls, crr_dict, amp_unit):
        """
        Fill holdcurr (i.e. holding current) field
        """
        hca = "holding_current"
        hcu = "holding_current_unit"

        holding_current = {}

        if hca and hcu in crr_dict:

            chca = crr_dict[hca]
            chcu = crr_dict[hcu]

            if chca and chcu:
                chca = chca[0] * manageConfig.conversion_factor(amp_unit, chcu)
                holding_current = {
                    "holdcurr": {
                        "value": [chca], "holdcurru": amp_unit,
                        "message": "Applied holding current: " + str(chca) +
                        " " + amp_unit
                    }
                }
        return holding_current


class manageConfig():
    """
    Manage configuration dictionary
    """

    # stimulus current units
    cu = ('a', 'da', 'ca', 'ma', 'ua', 'na', 'pa')

    # voltage unit
    vu = ('v', 'mv')

    # frequency unit
    fu = ('khz', 'hz')

    # all units
    all_units = cu + vu + fu

    @classmethod
    def conversion_factor(cls, unit_to, unit_from):
        """
        Extract conversion factors from unit_two to unit_one
        """

        conversion_table = {
            "v-mv": 1 / 1e3, "mv-v": 1e3, "a-da": 1 / 10, "a-ca": 1 / 1e2,
            "a-ma": 1 / 1e3, "a-ua": 1 / 1e6, "a-na": 1 / 1e9,
            "a-pa": 1 / 1e12, "da-a": 1 / 10, "da-ca": 10, "da-ma": 1e2,
            "da-ua": 1e5, "da-na": 1e8, "da-pa": 1e11, "ca-a": 1 / 1e2,
            "ca-da": 1 / 10, "ca-ma": 10, "ca-ua": 1e4, "ca-na": 1e7,
            "ca-pa": 1e10, "na-a": 1e9, "na-da": 1e8, "na-ca": 1e7,
            "na-ma": 1e6, "na-ua": 1e3, "na-pa": 1 / 1e3, "ma-a": 1e3,
            "ma-da": 1e2, "ma-ca": 10, "ma-ua": 1 / 1e3, "ma-na": 1e6,
            "ma-pa": 1e9, "ua-a": 1 / 1e6, "ua-da": 1 / 1e5, "ua-ca": 1 / 1e4,
            "ua-ma": 1 / 1e3, "ua-na": 1e3, "ua-pa": 1e6, "pa-a": 1e12,
            "pa-da": 1e11, "pa-ca": 1e10, "pa-ma": 1e9, "pa-ua": 1e6,
            "pa-na": 1e3, "khz-hz": 1 / 1e3, "hz-khz": 1e3,
        }

        conv_string = unit_to.lower() + "-" + unit_from.lower()

        if unit_to.lower() == unit_from.lower():
            return 1

        elif unit_to.lower() not in cls.all_units or \
                unit_from.lower() not in cls.all_units or \
                conv_string not in conversion_table:
            raise ValueError(
                "Given unit/s cannot be converted. Program is exiting")
        else:
            return conversion_table[conv_string]

    @classmethod
    def get_exclude_values(cls, cells_cellname, idx_file):
        """
        Extract stimulus values to be excluded and corresponding units from
        the configuration dictionary (i.e. 'config')
        """
        # extract stim traces to be excluded
        crr_exc = []
        crr_exc_u = []
        if "exclude" in cells_cellname and len(cells_cellname["exclude"]):

            exclude_values = cells_cellname["exclude"]

            if len(exclude_values) == 1:
                crr_exc = exclude_values[0]
                idx = 0
            elif len(exclude_values) - 1 >= idx_file:
                crr_exc = exclude_values[idx_file]
                idx = idx_file
            else:
                raise ValueError(
                    "'exclude' list length is shorter than " +
                    "the files list. Check your configuration. " +
                    "Program is exiting")

            if "exclude_unit" not in cells_cellname:
                crr_exc_u = 'nA'
            elif len(cells_cellname["exclude"]) == \
                    len(cells_cellname["exclude_unit"]):
                crr_exc_u = cells_cellname["exclude_unit"][idx]
            else:
                raise ValueError(
                    "'exclude' and 'exclude_unit' lists must " +
                    "have the same length. Program is exiting")

            conv_fact = manageConfig.conversion_factor('nA', crr_exc_u)
            crr_exc = [i * conv_fact for i in crr_exc]

        return [crr_exc, crr_exc_u]


class manageDicts():

    @classmethod
    def initialize_data_dict(cls):
        """
        Initialize the dictionary containing all data and info on the
        electrophysiological traces
        """

        data = OrderedDict()
        data['voltage'] = []
        data['current'] = []
        data['dt'] = []

        data['t'] = []
        data['ton'] = []
        data['toff'] = []
        data['tend'] = []
        data['amp'] = []
        data['hypamp'] = []
        data['filename'] = []

        return data

    @classmethod
    def fill_dict_single_trace(
            cls, data={}, voltage=0.0, current=0.0, dt=0.0, t=0.0, ton=0.0,
            toff=0.0, amp=0.0, hypamp=0.0, filename=""):

        data['voltage'].append(voltage)
        data['current'].append(current)
        data['dt'].append(dt)

        data['t'].append(t)
        data['tend'].append(t[-1])
        data['ton'].append(ton)
        data['toff'].append(toff)
        data['amp'].append(amp)
        data['hypamp'].append(hypamp)
        data['filename'].append(filename)

        return True



"""Plotting tools"""

"""
Copyright (c) 2020, EPFL/Blue Brain Project

 This file is part of BluePyEfe <https://github.com/BlueBrain/BluePyEfe>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import matplotlib
import matplotlib.colors as mplcol
import matplotlib.pyplot as plt
import colorsys
import numpy
import collections


def adjust_spines(ax, spines, color='k', d_out=10, d_down=[]):

    if d_down == []:
        d_down = d_out

    ax.set_frame_on(True)
    ax.patch.set_visible(False)

    for loc, spine in ax.spines.items():
        if loc in spines:
            if loc == 'bottom':
                spine.set_position(('outward', d_down))  # outward by 10 points
            else:
                spine.set_position(('outward', d_out))  # outward by 10 points
            # spine.set_smart_bounds(True)
        else:
            spine.set_visible(False)  # set_color('none') # don't draw spine

    # turn off ticks where there is no spine
    if 'left' in spines:
        ax.yaxis.set_ticks_position('left')

        if color != 'k':

            ax.spines['left'].set_color(color)
            ax.yaxis.label.set_color(color)
            ax.tick_params(axis='y', colors=color)

    elif 'right' not in spines:
        # no yaxis ticks
        ax.yaxis.set_ticks([])

    if 'right' in spines:
        ax.yaxis.set_ticks_position('right')

        if color != 'k':

            ax.spines['right'].set_color(color)
            ax.yaxis.label.set_color(color)
            ax.tick_params(axis='y', colors=color)

    if 'bottom' in spines:
        pass
        ax.xaxis.set_ticks_position('bottom')
        # ax.axes.get_xaxis().set_visible(True)

    else:
        # no xaxis ticks
        ax.xaxis.set_ticks([])
        ax.axes.get_xaxis().set_visible(False)


def light_palette(color, n_colors=6, reverse=False, lumlight=0.8, light=None):

    rgb = mplcol.colorConverter.to_rgb(color)

    if light is not None:
        light = mplcol.colorConverter.to_rgb(light)
    else:
        vals = list(colorsys.rgb_to_hls(*rgb))
        vals[1] = lumlight
        light = colorsys.hls_to_rgb(*vals)

    colors = [color, light] if reverse else [light, color]
    pal = mplcol.LinearSegmentedColormap.from_list("blend", colors)

    palette = pal(numpy.linspace(0, 1, n_colors))

    return palette


def tiled_figure(
        figname='', frames=1, columns=2, figs=collections.OrderedDict(),
        axs=None, orientation='page', width_ratios=None, height_ratios=None,
        top=0.97, bottom=0.05, left=0.07, right=0.97, hspace=0.6, wspace=0.2,
        dirname=''):

    if figname not in figs.keys():

        if axs is None:
            axs = []

        if orientation == 'landscape':
            figsize = (297 / 25.4, 210 / 25.4)
        elif orientation == 'page':
            figsize = (210 / 25.4, 297 / 25.4)

        params = {'backend': 'ps',
                  'axes.labelsize': 6,
                  'axes.linewidth': 0.5,
                  'font.size': 8,
                  'axes.titlesize': 8,
                  'legend.fontsize': 8,
                  'xtick.labelsize': 6,
                  'ytick.labelsize': 6,
                  'legend.borderpad': 0.2,
                  'legend.loc': 'best',
                  'text.usetex': False,
                  # 'pdf.fonttype': 42,
                  'figure.figsize': figsize}
        matplotlib.rcParams.update(params)

        fig = plt.figure(figname, facecolor='white')
        figs[figname] = {}
        figs[figname]['fig'] = fig
        figs[figname]['dirname'] = dirname

        if width_ratios is None:
            width_ratios = [1] * columns

        rows = int(numpy.ceil(frames / float(columns)))
        if height_ratios is None:
            height_ratios = [1] * rows

        gs = matplotlib.gridspec.GridSpec(
            rows, columns, height_ratios=height_ratios,
            width_ratios=width_ratios)
        gs.update(
            top=top,
            bottom=bottom,
            left=left,
            right=right,
            hspace=hspace,
            wspace=wspace)

        for fi in range(frames):
            axs.append(fig.add_subplot(
                gs[int(fi / columns), int(fi % columns)]))
            adjust_spines(axs[-1], ['left', 'bottom'], d_out=0)

        figs[figname]['axs'] = axs

    else:
        axs = figs[figname]['axs']

    return axs



"""Igor reader"""

"""
Copyright (c) 2020, EPFL/Blue Brain Project

 This file is part of BluePyEfe <https://github.com/BlueBrain/BluePyEfe>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import numpy
from . import igorpy
from collections import OrderedDict

import logging

logger = logging.getLogger(__name__)


def process(config=None,
            filename=None,
            cellname=None,
            expname=None,
            stim_feats=None,
            idx_file=None,
            ljp=0, v_corr=0):
    """Read recordings from an igor file"""

    path = config['path']
    cells = config['cells']
    features = config['features']
    options = config['options']

    data = OrderedDict()
    data['voltage'] = []
    data['current'] = []
    data['dt'] = []

    data['t'] = []
    data['ton'] = []
    data['toff'] = []
    data['tend'] = []
    data['amp'] = []
    data['hypamp'] = []
    data['filename'] = []

    ordinal = filename['ordinal']
    logger.debug(" Adding igor file with ordinal %s", ordinal)

    i_file = path + filename['i_file']
    v_file = path + filename['v_file']

    notes, wave = igorpy.read(v_file)

    if 'v_unit' not in filename:
        v_unit = notes.xUnits
    else:
        v_unit = filename['v_unit']

    if v_unit == 'V':
        v = wave * 1e3  # mV
    elif v_unit == 'mV':
        v = wave  # mV
    else:
        raise Exception(
            "Unit voltage not configured!")

    if 'dt' in filename:
        dt = filename['dt']
        if numpy.isclose(dt, notes.dx) is False:
            raise Exception(
                "Given stepsize %f does not match stepsize from "
                "wavenotes %f" % (dt, notes.dt))
    else:
        dt = notes.dx

    if 't_unit' not in filename:
        t_unit = notes.dUnits
        filename['t_unit'] = t_unit
    else:
        t_unit = filename['t_unit']

    if (t_unit == ""):
        filename['t_unit'] = "s"

    if (t_unit == "") or (t_unit == "s"):
        dt = dt * 1e3  # convert to ms
    t = dt * numpy.arange(0, len(wave))

    notes, wave = igorpy.read(i_file)

    if 'i_unit' not in filename:
        i_unit = notes.xUnits
    else:
        i_unit = filename['i_unit']

    if i_unit == 'A':
        i = wave * 1e9  # nA
    elif i_unit == 'pA':
        i = wave * 1e-3  # nA
    else:
        raise Exception(
            "Unit current not configured!")

    ton = options['onoff'][expname][0]
    toff = options['onoff'][expname][1]

    ion = int(ton / dt)

    if toff:
        ioff = int(toff / dt)
    else:
        ioff = False

    hypamp = numpy.mean(i[0:ion])  # estimate hyperpolarization current
    iborder = int((ioff - ion) * 0.1)  # 10% distance to measure step current
    # depolarization amplitude starting from hypamp!!

    if expname in ['APThreshold']:
        imax = numpy.argmax(i)
        toff = t[imax]
        trun = toff - ton
        ampoff = numpy.mean(i[int(imax - 10. / dt):imax]) - hypamp
        # extrapolate to get expected amplitude at 1 sec
        amp = ampoff / trun * 2000.
        # amp = ampoff
    elif expname in ['H40S8']:
        amp = numpy.mean(i[i > 0.1])
    elif "pulseAmp" in filename:
        if i_unit.lower() == "a":
            amp = filename["pulseAmp"] * 1e9
        elif i_unit.lower() == "pa":
            amp = filename["pulseAmp"] * 1e-3
        else:
            amp = filename["pulseAmp"]
    else:
        amp = numpy.mean(i[ion + iborder:ioff - iborder]) - hypamp

    # clean voltage from transients
    if expname in ['IDRest', 'IDrest', 'IDthresh', 'IDdepol']:
        cut_start = int(ion + numpy.ceil(1.0 / dt))
        v[ion:cut_start] = v[cut_start]
        cut_end0 = int(ioff - numpy.ceil(0.5 / dt))
        cut_end1 = int(ioff + numpy.ceil(2.0 / dt))
        v[cut_end0:cut_end1] = v[cut_end1]

    # delete second pulse
    elif expname in ['SpikeRec']:
        t = t[:int(50. / dt)]
        v = v[:int(50. / dt)]
        i = i[:int(50. / dt)]

    # normalize membrane potential to known value
    if v_corr:
        v = v - numpy.mean(v[0:ion]) + v_corr

    v = v - ljp  # correct junction potential

    data['voltage'].append(v)
    data['current'].append(i)
    data['dt'].append(dt)

    data['t'].append(t)
    data['tend'].append(t[-1])
    data['ton'].append(ton)
    data['toff'].append(toff)
    data['amp'].append(amp)
    data['hypamp'].append(hypamp)
    data['filename'].append(ordinal)

    logger.debug(" Added igor file with ordinal %s", ordinal)

    return data



"""ibf json reader"""

"""
Copyright (c) 2020, EPFL/Blue Brain Project

 This file is part of BluePyEfe <https://github.com/BlueBrain/BluePyEfe>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import numpy
import os
import json

from collections import OrderedDict

import logging

logger = logging.getLogger(__name__)


def process(config=None,
            filename=None,
            cellname=None,
            expname=None,
            stim_feats=None,
            idx_file=None,
            ljp=0, v_corr=0):
    """Read recordings from a json file"""

    path = config['path']
    cells = config['cells']
    features = config['features']
    options = config['options']

    data = OrderedDict()
    data['voltage'] = []
    data['current'] = []
    data['dt'] = []

    data['t'] = []
    data['ton'] = []
    data['toff'] = []
    data['tend'] = []
    data['amp'] = []
    data['hypamp'] = []
    data['filename'] = []

    # read stimulus features if present
    stim_feats = []
    if 'stim_feats' in cells[cellname]['experiments'][expname]:
        stim_feats = cells[cellname]['experiments'][expname]['stim_feats']

    logger.debug(" Adding ibf_json file %s", filename)

    f = path + os.sep + cellname + os.sep + filename + '.json'
    f = path + os.sep + filename + '.json'
    with open(f, 'r') as crr_file:
        f_data = json.load(crr_file)

    sampling_rate = f_data['sampling_rate']

    dt = 1. / int(sampling_rate) * 1e3

    # for all segments in file
    for idx, value in f_data['traces'].items():
        crr_amp = idx
        voltage = numpy.array(value).astype(numpy.float64)
        t = numpy.arange(len(voltage)) * dt

        ton = f_data['tonoff'][idx]['ton'][0]
        toff = f_data['tonoff'][idx]['toff'][0]
        ion = int(ton / dt)
        ioff = int(toff / dt)
        amp = numpy.float64(idx)

        current = []
        current = numpy.zeros(len(voltage))
        current[ion:ioff] = amp

        # estimate hyperpolarization current
        hypamp = numpy.mean(current[0:ion])

        # 10% distance to measure step current
        iborder = int((ioff - ion) * 0.1)

        # depolarization amplitude
        # amp = numpy.mean( current[ion+iborder:ioff-iborder] )
        voltage_dirty = voltage[:]

        # clean voltage from transients
        voltage[ion:ion + int(numpy.ceil(0.4 / dt))] = voltage[
            ion + int(numpy.ceil(0.4 / dt))]
        voltage[ioff:ioff + int(numpy.ceil(0.4 / dt))] = voltage[
            ioff + int(numpy.ceil(0.4 / dt))]

        # normalize membrane potential to known value (given in UCL
        # excel sheet)
        if v_corr:
            if len(v_corr) == 1 and v_corr[0] != 0.0:
                voltage = voltage - numpy.mean(voltage[0:ion]) + v_corr[0]
            elif len(v_corr) - 1 >= idx_file and v_corr[idx_file] != 0.0:
                voltage = voltage - numpy.mean(voltage[0:ion]) + v_corr[
                    idx_file]

        voltage = voltage - ljp

        # clip spikes after stimulus so they are not analysed
        voltage[ioff:] = numpy.clip(voltage[ioff:], -300, -40)

        if ('exclude' in cells[cellname] and
                any(abs(cells[cellname]['exclude'][idx_file] - amp) < 1e-4)):
            continue  # llb

        else:
            data['voltage'].append(voltage)
            data['current'].append(current)
            data['dt'].append(dt)

            data['t'].append(t)
            data['tend'].append(t[-1])
            data['ton'].append(ton)
            data['toff'].append(toff)
            data['amp'].append(amp)
            data['hypamp'].append(hypamp)
            data['filename'].append(filename)

    return data



import os
import numpy
import json


class printFeatures:
    TT_CONFIG_PATH = 'configs'
    TT_CONFIG_FILE = 'tabletools_config.json'

    @classmethod
    def dump_features(
            cls, all_feat_filename="all_features.txt", cellname="CELLNAME",
            trace_filename="", features_name=[], fel_vals=[], multvalnum=5,
            metadata={}, amp=0, stim_start=0, stim_end=0):
        counter = 0
        param_file = os.path.join(
            os.path.dirname(__file__), cls.TT_CONFIG_PATH,
            cls.TT_CONFIG_FILE)
        with open(param_file, 'r') as f:
            params = json.load(f)
        f.close()

        CELLINFO = params["CELLINFO"]
        MULTVALFEAT = params["MULTVALFEAT"]

        # create file with headers if needed
        if not os.path.exists(all_feat_filename):
            headers = []
            for i in CELLINFO:
                headers.append(i)
            headers.append('stim')
            headers.append('stim_start')
            headers.append('stim_end')
            for i in features_name:
                if i in MULTVALFEAT:
                    for j in range(multvalnum - 1):
                        headers.append(i + '___event' + str(j + 1))
                    headers.append(i + '___eventlast')
                else:
                    headers.append(i)
            headers.append('\n')
            with open(all_feat_filename, 'w') as f:
                f.write('\t'.join(headers))
        f = open(all_feat_filename, 'a')
        crr_info_str = []
        crr_sweep_str = []
        for i in CELLINFO:
            if i == 'cell_id':
                crr_info_str.append(str(cellname))
            elif i == 'filename':
                crr_info_str.append(str(trace_filename))
            else:
                crr_info_str.append(metadata[i])

        # append stim info
        crr_info_str.append(str(amp))
        crr_info_str.append(str(stim_start))
        crr_info_str.append(str(stim_end))

        # convert to string
        crr_sweep_str = list(crr_info_str)

        # for every extracted feature
        for ii in features_name:
            crr_feature = fel_vals[0][ii]
            if ii not in MULTVALFEAT:
                if crr_feature is not None and not \
                        (
                        numpy.array_equal(
                            crr_feature,
                            numpy.array([]))):
                    crr_sweep_str.append(str(crr_feature[0]))
                else:
                    crr_sweep_str.append(str(numpy.nan))
            else:
                crr_feat_val = [numpy.nan] * multvalnum
                if crr_feature is not None and len(crr_feature) > 0:
                    crr_feat_len = len(crr_feature)
                    for y in range(multvalnum - 2):
                        if len(crr_feature) > y + 1:
                            crr_feat_val[y + 1] = crr_feature[y + 1]
                    if any(crr_feature):
                        crr_feat_val[0] = crr_feature[0]
                        crr_feat_val[-1] = crr_feature[-1]

                # append values to the string to be printed
                for w in crr_feat_val:
                    crr_sweep_str.append(str(w))

        crr_final_str = '\t'.join(crr_sweep_str)
        crr_final_str += "\n"
        f.write(crr_final_str)



"""Axon reader"""

"""
Copyright (c) 2020, EPFL/Blue Brain Project

 This file is part of BluePyEfe <https://github.com/BlueBrain/BluePyEfe>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from neo import io
from collections import OrderedDict
import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
import numpy
import os
import sys
import neo.rawio.axonrawio
import json
from . import common


def process(config=None,
            filename=None,
            cellname=None,
            expname=None,
            stim_feats=None,
            idx_file=None,
            ljp=0, v_corr=0):
    """Read recordings from an axon file"""
    path = config['path']
    cells = config['cells']
    # features = config['features']
    # options = config['options']

    # initialize data dictionary
    data = common.manageDicts.initialize_data_dict()

    # read metadata file if present
    meta_dict = {}

    f_meta = os.path.join(path, cellname, filename + '_metadata.json')
    if os.path.exists(f_meta):
        with open(f_meta) as json_metadata:
            meta_dict = json.load(json_metadata)
        json_metadata.close()

    logger.info(" Adding axon file %s", filename)

    f = os.path.join(path, cellname, filename + '.abf')
    r = io.AxonIO(filename=f)

    # read header

    # Below line doesn't work anymore due to api change
    # Now using rawio
    # header = r.read_header()

    header = neo.rawio.axonrawio.parse_axon_soup(f)

    # get number of episodes from header
    nbepisod = get_nbepisod(header)

    # read sampling rate
    # sampling_rate = 1.e6 / header['protocol']['fADCSequenceInterval']
    # dt = 1. / int(sampling_rate) * 1e3
    # version = header['fFileVersionNumber']

    # read data block
    bl = r.read_block(lazy=False)

    # initialize all stimuli data structure
    all_stims = []

    stim_feats = None
    res = []
    stim_info_flag = False

    if 'stim_info' in cells[cellname]['experiments'][expname] and \
            'stim_feats' in cells[cellname]['experiments'][expname]:
        raise Exception(
            "Both 'stim_info' and 'stim_feats' are given in" +
            " the configuration parameters. Only one or none is allowed.")

    # read stimulus info from the option stim_feats if any
    if 'stim_feats' in cells[cellname]['experiments'][expname]:
        stim_feats_crr = cells[cellname]['experiments'][expname]['stim_feats']

        # extract stimulus info from the config file
        if isinstance(stim_feats_crr, dict) is True:
            stim_feats = stim_feats_crr
        elif isinstance(stim_feats_crr, list) is True:
            if len(stim_feats_crr) == 1:
                stim_feats = stim_feats_crr[0]
            elif len(stim_feats_crr) == len(cells[cellname][
                    'experiments'][expname]['files']):
                stim_feats = stim_feats_crr[idx_file]
            else:
                raise ValueError(
                    "The 'stim_feats' list must have length " +
                    "equals to 1 or to the length of the 'files' list." +
                    "Please, check your configuration")
        else:
            raise ValueError(
                "'stim_feats' must be a list or a dictionary. Please," +
                " check your configuration")

        stim_feats['filename'] = filename
        res = common.manageMetadata.stim_feats_from_meta(
            stim_feats, nbepisod)

        logger.info("File: %s. Found info in config file", filename)

    # extract stim from metadata file if any
    elif 'stim_info' in cells[cellname]['experiments'][expname] and \
            cells[cellname]['experiments'][expname]['stim_info'] is not None:
        stim_info = cells[cellname]['experiments'][expname]['stim_info']
        stim_info_flag = True
        try:
            sampling_rate = 1.e6 / header['protocol']['fADCSequenceInterval']
        except Exception as e:
            logger.info(
                "Unable to find recording frequency in file: %s." +
                "The ABF file version is probably older than v2", filename)

    else:
        # read metadata file if present
        stim_feats = {}

        f_meta = os.path.join(path, cellname, filename + '_metadata.json')
        if os.path.exists(f_meta):
            with open(f_meta) as json_metadata:
                stim_feats = json.load(json_metadata)
            json_metadata.close()

        # if metadata with stimulus info could be extracted
        if stim_feats:
            res = common.manageMetadata.stim_feats_from_meta(
                meta_dict, nbepisod)
            logger.info("File: %s. Found info in metadata file", filename)

    # if stimulus has been extracted from stim_feats or metadata file
    if res and res[0]:
        # extract sampling rate from metadata
        sampling_rate = res[1]["r"][0]
        all_stims = res[1]
    elif not stim_info_flag:
        # extract stim from header if any
        logger.info(
            "File: %s. No stimulus info found in config or " +
            "metadata file. Extracting info from the file header.",
            filename)
        res = stim_feats_from_header(header)
        all_stims = res[1]

        # extract sampling rate from header
        sampling_rate = sampling_rate_from_header(header)[1]["r"]

    # if no stimulus could be extracted
    if not all_stims and not stim_info_flag:
        raise ValueError("No valid stimulus was found in metadata or files. \
            Skipping current file")

    dt = 1. / int(sampling_rate) * 1e3

    # for all segments in file
    for i_seg, seg in enumerate(bl.segments):

        if stim_info_flag:
            voltage = numpy.array(
                seg.analogsignals[0]).astype(
                numpy.float64).flatten()
            current = numpy.array(
                seg.analogsignals[1]).astype(
                numpy.float64).flatten()
            t = numpy.arange(len(voltage)) * dt

            ton = stim_info['ton']
            toff = stim_info['toff']
            ion = int(ton / dt)
            ioff = int(toff / dt)

            if 'tamp' in stim_info:
                tamp = [
                    int(stim_info['tamp'][0] / dt),
                    int(stim_info['tamp'][1] / dt)]
            else:
                tamp = [ion, ioff]

            i_unit = stim_info['i_unit']

            if i_unit.lower() == 'a':
                current = current * 1e9  # nA
            elif i_unit.lower() == 'pa':
                current = current * 1e-3  # nA
            elif i_unit.lower() == 'na':
                pass
            else:
                raise Exception(
                    "Unit current not configured!")

            amp = numpy.nanmean(current[tamp[0]:tamp[1]])
            hypamp = numpy.nanmean(current[0:ion])

        else:
            # the following loop is needed because the voltage is not always
            # in the first array of the segment
            voltage = []
            for i_asig, asig in enumerate(seg.analogsignals):
                crr_unit = str(seg.analogsignals[i_asig].units.dimensionality)
                if str(crr_unit.lower()) in common.manageConfig.vu:
                    voltage = numpy.array(asig).astype(numpy.float64).flatten()

            if len(voltage) == 0:
                continue

            t = numpy.arange(len(voltage)) * dt
            ton = all_stims["st"][i_seg]
            toff = all_stims["en"][i_seg]
            ion = int(ton / dt)
            ioff = int(toff / dt)
            amp = numpy.float64(all_stims["crr_val"][i_seg])
            stim_u = all_stims["u"][i_seg]

            # convert stimulus amplitude to nA
            conv_fact = common.manageConfig.conversion_factor('nA', stim_u)

            amp = amp * conv_fact
            if conv_fact != 1:
                stim_u = "nA"

            current = []
            current = numpy.zeros(len(voltage))
            current[ion:ioff] = amp

            # estimate hyperpolarization current
            hypamp = numpy.mean(current[0:ion])

            # 10% distance to measure step current
            iborder = int((ioff - ion) * 0.1)

            # clean voltage from transients
            voltage[ion:ion + int(numpy.ceil(0.4 / dt))] = \
                voltage[ion + int(numpy.ceil(0.4 / dt))]
            voltage[ioff:ioff + int(numpy.ceil(0.4 / dt))] = \
                voltage[ioff + int(numpy.ceil(0.4 / dt))]

        # normalize membrane potential to known value
        # (given in UCL excel sheet)
        if isinstance(v_corr, list):
            if len(v_corr) == 1 and v_corr[0] != 0.0:
                voltage = voltage - numpy.mean(voltage[0:ion]) + v_corr[0]
            elif len(v_corr) - 1 >= idx_file and v_corr[0] != 0.0:
                voltage = voltage - numpy.mean(voltage[0:ion]) \
                    + v_corr[idx_file]
        elif v_corr != 0.0:
            voltage = voltage - numpy.mean(voltage[0:ion]) + v_corr

        voltage = voltage - ljp

        # clip spikes after stimulus so they are not analysed
        voltage[ioff:] = numpy.clip(voltage[ioff:], -300, -40)

        # extract stim traces to be excluded
        [crr_exc, crr_exc_u] = common.manageConfig.get_exclude_values(
            cells[cellname], idx_file)

        if not len(crr_exc) == 0 and any(abs(crr_exc - amp) < 1e-4):
            continue  # llb
        else:
            common.manageDicts.fill_dict_single_trace(
                data=data, voltage=voltage, current=current, dt=dt, t=t,
                ton=ton, toff=toff, amp=amp, hypamp=hypamp,
                filename=filename)
    resp_check = check_validity(data)

    return data


def stim_feats_from_header(header):
    """Get stimulus features from the file header (author: Luca Leonardo
    Bologna)"""
    # read sampling rate
    sampling_rate = 1.e6 / header['protocol']['fADCSequenceInterval']

    # read file version
    version = header['fFileVersionNumber']

    # extract protocol for version >=.2
    if version >= 2.:
        # prot = r.read_protocol() # read protocol
        # read info for DAC
        dictEpochInfoPerDAC = header['dictEpochInfoPerDAC']

        # if field is empty
        if not (dictEpochInfoPerDAC):
            return (0, "No 'dictEpochInfoPerDAC' field")

        # if field is not empty, read all stimulus segments
        else:
            valid_epoch_dicts = [k for k, v in dictEpochInfoPerDAC.items()
                                 if bool(v)]

            # if more than one channel is activated for the stimulus
            # or a number of epochs different than 3 is found
            if len(valid_epoch_dicts) != 1 or len(dictEpochInfoPerDAC[0]) != 3:
                return (0, 'Exiting. More than one channel used \
                        for stimulation')
            else:
                # read all stimulus epochs

                # TODO IS THIS CORRECT ?
                # being here, len(valid_epoch_dicts) == 1, so all the stimulus
                # epochs should be contained in its first and only element
                # (normally zero)

                k = valid_epoch_dicts[0]

                stim_epochs = dictEpochInfoPerDAC[valid_epoch_dicts[0]]

                # read enabled waveforms
                stim_ch_info = [(i['DACChNames'], i['DACChUnits'],
                                 i['nDACNum']) for i in header['listDACInfo']
                                if bool(i['nWaveformEnable'])]

                # if epoch initial levels and increment are not
                # compatible with a step stimulus
                if (stim_epochs[0]['fEpochInitLevel'] !=
                        stim_epochs[2]['fEpochInitLevel'] or
                        stim_epochs[0]['fEpochLevelInc'] !=
                        stim_epochs[2]['fEpochLevelInc'] or
                        float(
                            format(
                                stim_epochs[0][
                                    'fEpochLevelInc'], '.3f')) != 0 or
                        (len(stim_ch_info) != 1 or
                            stim_ch_info[0][2] != k)):
                    # return 0 with message
                    return (
                        0, "A stimulus different from the steps has been \
                        detected")
                else:
                    ty = "step"
                    u = stim_ch_info[0][1]
                    # number of ADC channels
                    nADC = header['sections']['ADCSection']['llNumEntries']
                    # number of samples per episode
                    nSam = header['protocol']['lNumSamplesPerEpisode'] / nADC
                    # number of actual episodes
                    nEpi = header['lActualEpisodes']

                    # read first stimulus epoch
                    e_zero = header['dictEpochInfoPerDAC'][
                        stim_ch_info[0][2]][0]
                    # read second stimulus epoch
                    e_one = header['dictEpochInfoPerDAC'][
                        stim_ch_info[0][2]][1]
                    # index of stimulus beginning
                    i_last = int(nSam * 15625 / 10**6)
                    # create array for all stimulus info
                    all_stim_feats = {
                        "ty": [],
                        "st": [],
                        "en": [],
                        "crr_val": [],
                        "u": []
                    }

                    # step increment
                    e_one_inc = float(format(e_one['fEpochLevelInc'],
                                             '.3f'))
                    # step initial level
                    e_one_init_level = float(format(e_one['fEpochInitLevel'],
                                                    '.3f'))

                    # for every episode, compute stimulus start, stimulus end,
                    # stimulus value
                    for epiNum in range(nEpi):
                        st = i_last + e_zero['lEpochInitDuration'] + \
                            e_zero['lEpochDurationInc'] * epiNum
                        en = st + e_one['lEpochInitDuration'] +  \
                            e_one['lEpochDurationInc'] * epiNum
                        crr_val_full = float(format(e_one_init_level +
                                                    e_one_inc * epiNum, '.3f'))
                        crr_val = float(format(crr_val_full, '.3f'))
                        st = 1 / sampling_rate * st * 1e3
                        en = 1 / sampling_rate * en * 1e3

                        # convert unit from bytes to string
                        if sys.version_info[0] >= 3:
                            ustr = str(u, 'utf-8')
                        else:
                            ustr = str(u)

                        all_stim_feats["ty"].append(ty)
                        all_stim_feats["st"].append(st)
                        all_stim_feats["en"].append(en)
                        all_stim_feats["crr_val"].append(crr_val)
                        all_stim_feats["u"].append(ustr)

                    return (1, all_stim_feats)
    else:
        return (0, {})


def sampling_rate_from_header(header):
    """
    Extract sampling rate from the header of the abf.file
    """

    # read version
    version = header['fFileVersionNumber']  # read file version

    if version < 2.:
        # read sampling rate
        nbchannel = header['nADCNumChannels']
        sampling_rate = 1. / (header['fADCSampleInterval'] * nbchannel * 1.e-6)

    elif version >= 2.:
        # read sampling rate
        sampling_rate = 1.e6 / header['protocol']['fADCSequenceInterval']

    return(1, {"r": sampling_rate, "ru": "Hz"})


def get_nbepisod(header):
    # read version
    version = header['fFileVersionNumber']  # read file version

    if version < 2.:
        # read sampling rate
        nbepisod = header['lActualEpisodes']

    elif version >= 2.:
        # read sampling rate
        nbepisod = header['lActualEpisodes']

    return nbepisod


#
def check_validity(data):
    # extract number of traces
    nb_traces = len(data["voltage"])

    # extract number of stimuli
    nb_stims = len(data['current'])

    if nb_traces != nb_stims:
        raise Exception("Number of traces and number of given \
                stimuli are different for file: " + filename)
    else:
        return (1, "")



"""Extractor class"""

"""
Copyright (c) 2020, EPFL/Blue Brain Project

 This file is part of BluePyEfe <https://github.com/BlueBrain/BluePyEfe>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

import matplotlib
matplotlib.use('Agg')  # noqa
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages

import numpy
import sys
import efel
import os
import logging
import gzip
import json
import pprint

from itertools import cycle
from collections import OrderedDict


from . import tools
from .tools import tabletools
from . import plottools
from . import extra
from .formats import common


logger = logging.getLogger(__name__)
logging.basicConfig(stream=sys.stdout)
logger = logging.getLogger()


class Extractor(object):

    """Extractor class"""

    def __init__(self, mainname='PC', config=OrderedDict()):
        """Constructor

        Args:
            mainname (str): name of the cell type. Used to create the output
            directory
            config (dict): metadata containing the protocols and cells for
            which to extract the efeatures.
        """

        self.config = config
        self.path = config['path']
        self.cells = config['cells']
        self.features = config['features']
        self.options = config['options']

        for experiment in self.features:
            f = self.features[experiment]
            self.features[experiment] = sorted(
                set(f), key=lambda x: f.index(x))

        self.format = config['format']

        self.dataset = OrderedDict()
        self.dataset_mean = OrderedDict()

        self.max_per_plot = 16

        conv_fact = 1

        if "relative" not in self.options:
            self.options["relative"] = False

        if "amp_min" not in self.options:
            self.options["amp_min"] = 0.001  # minimum current amplitude used

        if "peak_min" not in self.options:
            # minimum current amplitude used for spike detection
            self.options["peak_min"] = 0.001

        if "target" not in self.options:
            self.options["target"] = [100., 150., 200., 250.]
        else:
            if "target_unit" in self.options:
                target_unit = self.options["target_unit"].lower()
                conv_fact = common.manageConfig.conversion_factor(
                    'nA', target_unit)
                if conv_fact != 1:
                    conv_target = [
                        i * conv_fact for i in
                        self.options["target"]]
                    self.options["target"] = conv_target

        if "tolerance" not in self.options:
            self.options["tolerance"] = 10

        if "strict_stiminterval" not in self.options:
            self.options["strict_stiminterval"] = {'base': False}

        if isinstance(self.options["tolerance"], list) is False:
            if conv_fact != 1:
                tolerance = self.options["tolerance"]
                self.options["tolerance"] = conv_fact * tolerance
            self.options["tolerance"] =\
                numpy.ones(len(self.options["target"]))\
                * self.options["tolerance"]
        elif conv_fact != 1:
            self.options["tolerance"] = \
                [x * conv_fact for x in self.options["tolerance"]]

        if "nanmean" not in self.options:
            self.options["nanmean"] = False

        if "nanmean_cell" not in self.options:
            self.options["nanmean_cell"] = True

        if "nangrace" not in self.options:
            self.options["nangrace"] = 0

        if "delay" not in self.options:
            self.options["delay"] = 0

        if "posttime" not in self.options:
            self.options["posttime"] = 200

        if "spike_threshold" not in self.options:
            self.options["spike_threshold"] = 2

        if "logging" not in self.options:
            self.options["logging"] = False

        if self.options["logging"]:
            logger.setLevel(logging.INFO)
        else:
            logger.setLevel(logging.ERROR)

        self.saveraw = False
        if ('saveraw' in self.options) and self.options['saveraw']:
            self.saveraw = True

        self.colors = OrderedDict()
        self.colors['b1'] = '#1F78B4'  # 377EB8
        self.colors['b2'] = '#A6CEE3'
        self.colors['g1'] = '#33A02C'  # 4DAF4A
        self.colors['g2'] = '#B2DF8A'
        self.colors['r1'] = '#E31A1C'  # E41A1C
        self.colors['r2'] = '#FB9A99'
        self.colors['o1'] = '#FF7F00'  # FF7F00
        self.colors['o2'] = '#FDBF6F'
        self.colors['p1'] = '#6A3D9A'  # 984EA3
        self.colors['p2'] = '#CAB2D6'

        self.colors['ye1'] = '#FFFF33'
        self.colors['br1'] = '#A65628'
        self.colors['pi1'] = '#F781BF'
        self.colors['gr1'] = '#999999'
        self.colors['k1'] = '#000000'

        self.markerlist = ['o', '*', '^', 'H', 'D', 's', 'p', '.', '8', '+']
        self.colorlist = [
            self.colors['b1'],
            self.colors['g1'],
            self.colors['r1'],
            self.colors['o1'],
            self.colors['p1'],
            self.colors['ye1']]
        self.experiments = []

        self.mainname = mainname
        self.maindirname = mainname + os.sep  # llb
        tools.makedir(self.maindirname)

        self.thresholds_per_cell = OrderedDict()
        self.hypamps_per_cell = OrderedDict()

        self.extra_features = ['spikerate_tau_jj', 'spikerate_drop',
                               'spikerate_tau_log', 'spikerate_tau_fit',
                               'spikerate_tau_slope']

    def newmeancell(self, a):
        if (self.options["nanmean_cell"] or
                (numpy.sum(numpy.isnan(a)) <= self.options["nangrace"])):
            return numpy.nanmean(a)
        else:
            return numpy.mean(a)

    def newstdcell(self, a):
        if (self.options["nanmean_cell"] or
                (numpy.sum(numpy.isnan(a)) <= self.options["nangrace"])):
            return numpy.nanstd(a)
        else:
            return numpy.std(a)

    def boxcoxcell(self, a, nanopt="nanmean_cell",
                   lm_vec=numpy.linspace(-3, 3, 41)):
        if ((self.options[nanopt] or
             (numpy.sum(numpy.isnan(a)) <= self.options["nangrace"])) and
                (len(a) > 0)):
            a = numpy.array(a)
            a = a[~numpy.isnan(a)]
        else:
            return float('NaN'), float('NaN'), float('NaN'), float('NaN')

        return float('NaN'), float('NaN'), float('NaN'), float('NaN')

    def newmean(self, a):
        if (self.options["nanmean"] or
                (numpy.sum(numpy.isnan(a)) <= self.options["nangrace"])):
            return numpy.nanmean(a)
        else:
            return numpy.mean(a)

    def newstd(self, a):
        if (self.options["nanmean"] or
                (numpy.sum(numpy.isnan(a)) <= self.options["nangrace"])):
            return numpy.nanstd(a)
        else:
            return numpy.std(a)

    def create_dataset(self):
        """Read the trace files and add them to the dictionnary self.dataset"""
        logger.info(" Filling dataset")

        for i_cell, cellname in enumerate(self.cells):

            self.dataset[cellname] = OrderedDict()

            v_corr = self.cells[cellname]['v_corr']
            self.dataset[cellname]['v_corr'] = v_corr

            if 'ljp' in self.cells[cellname]:
                ljp = self.cells[cellname]['ljp']
            else:
                ljp = 0
            self.dataset[cellname]['ljp'] = ljp

            dataset_cell_exp = OrderedDict()
            self.dataset[cellname]['experiments'] = dataset_cell_exp

            for i_exp, expname in enumerate(
                    self.cells[cellname]['experiments']):

                files = self.cells[cellname]['experiments'][expname]['files']

                # read stimulus features if present
                stim_feats = []
                if 'stim_feats' in self.cells[cellname]['experiments'][
                        expname]:
                    stim_feats = \
                        self.cells[cellname][
                            'experiments'][expname]['stim_feats']

                if len(files) > 0:
                    logger.debug(" Adding experiment %s", expname)

                    if expname not in self.experiments:
                        self.experiments.append(expname)

                    dataset_cell_exp[expname] = OrderedDict()

                    dataset_cell_exp[expname]['location'] = \
                        self.cells[cellname][
                        'experiments'][expname]['location']

                    dataset_cell_exp[expname]['voltage'] = []
                    dataset_cell_exp[expname]['current'] = []
                    dataset_cell_exp[expname]['dt'] = []
                    dataset_cell_exp[expname]['filename'] = []

                    dataset_cell_exp[expname]['t'] = []
                    dataset_cell_exp[expname]['ton'] = []
                    dataset_cell_exp[expname]['tend'] = []
                    dataset_cell_exp[expname]['toff'] = []
                    dataset_cell_exp[expname]['amp'] = []
                    dataset_cell_exp[expname]['hypamp'] = []
                    dataset_cell_exp[expname]['files'] = []
                    dataset_cell_exp[expname]['rawfiles'] = []

                    for idx_file, filename in enumerate(files):

                        data = self.process_file(config=self.config,
                                                 filename=filename,
                                                 cellname=cellname,
                                                 expname=expname,
                                                 stim_feats=stim_feats,
                                                 idx_file=idx_file,
                                                 v_corr=v_corr,
                                                 ljp=ljp)

                        dataset_cell_exp[expname]['voltage'] += data['voltage']
                        dataset_cell_exp[expname]['current'] += data['current']
                        dataset_cell_exp[expname]['dt'] += data['dt']
                        dataset_cell_exp[expname]['filename'] += \
                            data['filename']
                        dataset_cell_exp[expname]['rawfiles'].append(filename)

                        dataset_cell_exp[expname]['t'] += data['t']
                        dataset_cell_exp[expname]['ton'] += data['ton']
                        dataset_cell_exp[expname]['tend'] += data['tend']
                        dataset_cell_exp[expname]['toff'] += data['toff']
                        dataset_cell_exp[expname]['amp'] += data['amp']
                        dataset_cell_exp[expname]['hypamp'] += data['hypamp']

    def process_file(self, **kwargs):
        """Get data from a trace file"""
        if self.format == 'igor':
            from .formats import igor
            return igor.process(**kwargs)
        elif self.format == 'axon':
            from .formats import axon
            return axon.process(**kwargs)
        elif self.format == 'csv_lccr':
            from .formats import csv_lccr
            return csv_lccr.process(**kwargs)
        elif self.format == 'spike2':
            from .formats import spike2
            return spike2.process(**kwargs)
        elif self.format == 'ibf_json':
            from .formats import ibf_json
            return ibf_json.process(**kwargs)
        else:
            raise ValueError('Unrecognized trace format: %s' % self.format)

    def plt_traces(self):
        """Plot traces"""
        logger.info(" Plotting traces")

        for i_cell, cellname in enumerate(self.dataset):

            dirname = self.maindirname + cellname
            tools.makedir(dirname)
            dataset_cell_exp = self.dataset[cellname]['experiments']

            for i_exp, expname in enumerate(dataset_cell_exp):

                voltages = dataset_cell_exp[expname]['voltage']
                currents = dataset_cell_exp[expname]['current']
                amps = dataset_cell_exp[expname]['amp']
                ts = dataset_cell_exp[expname]['t']
                filenames = dataset_cell_exp[expname]['filename']

                colors = []

                colorcycler = cycle(self.colorlist)

                color_dict = {u: next(colorcycler)
                              for u in list(set(filenames))}
                colors = [color_dict[f] for f in filenames]

                isort = numpy.argsort(amps)
                amps = numpy.array(amps)[isort]
                ts = numpy.array(ts)[isort]
                voltages = numpy.array(voltages)[isort]
                currents = numpy.array(currents)[isort]
                colors = numpy.array(colors)[isort]
                filenames = numpy.array(filenames)[isort]

                n_plot = len(voltages)

                if n_plot <= self.max_per_plot:
                    frames = n_plot
                    n_fig = 1
                else:
                    frames = self.max_per_plot
                    n_fig = int(numpy.ceil(n_plot / float(self.max_per_plot)))

                axs = []
                figs = OrderedDict()

                axs_c = []
                figs_c = OrderedDict()

                for i_fig in range(n_fig):
                    figname = cellname.split(
                        '/')[-1] + "_" + expname + "_" + str(i_fig)
                    axs = plottools.tiled_figure(
                        figname, frames=frames, columns=2, figs=figs, axs=axs,
                        top=0.97, bottom=0.04, left=0.07, right=0.97,
                        hspace=0.75, wspace=0.2)

                    figname_c = cellname.split(
                        '/')[-1] + "_" + expname + "_" + str(i_fig) + "_i"
                    axs_c = plottools.tiled_figure(
                        figname_c, frames=frames, columns=2, figs=figs_c,
                        axs=axs_c, top=0.97, bottom=0.04, left=0.07,
                        right=0.97, hspace=0.75, wspace=0.2)

                for i_plot in range(n_plot):
                    # reduce figure title length if too long
                    filename_string = filenames[i_plot]
                    if len(filename_string) > 50:
                        filename_fin = filename_string[:18] + ' ... ' + \
                            filename_string[-18:]
                    else:
                        filename_fin = filename_string

                    axs[i_plot].plot(
                        ts[i_plot],
                        voltages[i_plot],
                        color=colors[i_plot],
                        clip_on=False)
                    axs[i_plot].set_title(
                        cellname + " " + expname + " amp:" +
                        str(amps[i_plot]) +
                        " file:" + filename_fin)

                    axs_c[i_plot].plot(
                        ts[i_plot],
                        currents[i_plot],
                        color=colors[i_plot],
                        clip_on=False)
                    axs_c[i_plot].set_title(
                        cellname + " " + expname + " amp:" +
                        str(amps[i_plot]) +
                        " file:" + filename_fin)

                # plt.show()

                for i_fig, figname in enumerate(figs):
                    fig = figs[figname]
                    fig['fig'].savefig(
                        dirname + '/' + figname + '.pdf', dpi=300)
                    plt.close(fig['fig'])

                for i_fig, figname in enumerate(figs_c):
                    fig = figs_c[figname]
                    fig['fig'].savefig(
                        dirname + '/' + figname + '.pdf', dpi=300)
                    plt.close(fig['fig'])

    def extract_features(self, threshold=-20):
        """Extract features from the traces"""
        logger.info(" Extracting features")

        efel.setThreshold(threshold)
        logger.info(" Setting spike threshold to %.2f mV", threshold)

        # if print_table flag is set, dump all extracted feature to a .csv file
        print_table_flag = False
        if 'print_table' in self.options and self.options[
                'print_table']['flag']:
            print_table_flag = True
            all_feat_filename = os.path.join(
                self.mainname, 'all_feature_table.txt')
            if os.path.exists(all_feat_filename):
                os.remove(all_feat_filename)
            if not hasattr(self, "metadataset"):
                self.create_metadataset()

        # set flag to convert zero value to nan for features set by the user
        if 'zero_to_nan' in self.options and \
                'flag' in self.options['zero_to_nan']:
            ZERO_TO_NAN = self.options['zero_to_nan']['flag']
        else:
            ZERO_TO_NAN = False

        for i_cell, cellname in enumerate(self.dataset):

            dataset_cell_exp = self.dataset[cellname]['experiments']
            for i_exp, expname in enumerate(dataset_cell_exp):

                if expname in self.options["strict_stiminterval"].keys():
                    strict_stiminterval = self.options["strict_stiminterval"][
                        expname]
                else:
                    strict_stiminterval = self.options["strict_stiminterval"][
                        'base']
                efel.setIntSetting("strict_stiminterval", strict_stiminterval)

                ts = dataset_cell_exp[expname]['t']
                voltages = dataset_cell_exp[expname]['voltage']
                tons = dataset_cell_exp[expname]['ton']
                toffs = dataset_cell_exp[expname]['toff']
                amps = dataset_cell_exp[expname]['amp']

                features_all = self.features[expname] + ['peak_time']

                if 'threshold' in self.cells[cellname]['experiments'][expname]:
                    threshold = self.cells[cellname][
                        'experiments'][expname]['threshold']
                    logger.info(" Setting threshold to %f", threshold)
                    efel.setThreshold(threshold)

                dataset_cell_exp[expname]['features'] = OrderedDict()
                for feature in features_all:
                    dataset_cell_exp[expname]['features'][feature] = []
                dataset_cell_exp[expname]['features']['numspikes'] = []

                # iterate over all voltages individually to get features
                # and correct them if needed
                for i_seg in range(len(voltages)):

                    trace = OrderedDict()
                    trace['T'] = ts[i_seg]
                    trace['V'] = voltages[i_seg]
                    trace['stim_start'] = [tons[i_seg]]
                    trace['stim_end'] = [toffs[i_seg]]
                    traces = [trace]
                    amp = amps[i_seg]

                    efel.setDoubleSetting('stimulus_current', amp)

                    features_all_ = [
                        f for f in features_all
                        if f not in self.extra_features]

                    fel_vals = efel.getFeatureValues(
                        traces, features_all_, raise_warnings=False)

                    peak_times = fel_vals[0]['peak_time']

                    for feature in features_all:

                        if feature == 'peak_time':
                            if (len(peak_times) >
                                    0) and amp >= self.options["peak_min"]:
                                numspike = len(peak_times)
                            else:
                                numspike = 0

                        elif feature == 'spikerate_tau_jj':
                            if len(peak_times) > 4:
                                f = extra.spikerate_tau_jj(peak_times)
                            else:
                                f = None

                        elif feature == 'spikerate_drop':
                            if len(peak_times) > 4:
                                f = extra.spikerate_drop(peak_times)
                            else:
                                f = None

                        elif feature == 'spikerate_tau_log':
                            if len(peak_times) > 4:
                                f = extra.spikerate_tau_log(peak_times)
                            else:
                                f = None

                        elif feature == 'spikerate_tau_fit':
                            if len(peak_times) > 4:
                                f = extra.spikerate_tau_fit(peak_times)
                            else:
                                f = None

                        elif feature == 'spikerate_tau_slope':
                            if len(peak_times) > 4:
                                f = extra.spikerate_tau_slope(peak_times)
                            else:
                                f = None
                        elif fel_vals[0][feature] is not None and \
                                len(fel_vals[0][feature]) == 1 and \
                                fel_vals[0][feature][0] == 0 and ZERO_TO_NAN \
                                and feature in self.options[
                                    'zero_to_nan'][
                                    'mean_features_no_zeros']:
                            if self.options[
                                    'zero_to_nan']['value'] == 'stim_end':
                                fel_vals[0][feature] = [toffs[i_seg]]
                                f = [toffs[i_seg]]
                            elif self.options[
                                    'zero_to_nan']['value'] == 'nan':
                                fel_vals[0][feature] = None
                                f = None
                            else:
                                logger.info(
                                    "Unrecognized value " +
                                    "for zero_to_nan option")
                        else:
                            f = fel_vals[0][feature]

                        if abs(amp) < self.options["amp_min"]:
                            f = float('nan')
                        elif f is not None:
                            f = numpy.mean(f)
                        else:
                            f = float('nan')

                        if "trace_check" in self.options and \
                                self.options["trace_check"] is False:
                            pass
                        else:
                            # exclude any activity outside stimulus (20 ms
                            # grace period)
                            if (any(numpy.atleast_1d(peak_times) <
                                    trace['stim_start'][0]) or any(
                                    numpy.atleast_1d(peak_times) >
                                    trace['stim_end'][0] + 20)):
                                f = float('nan')

                        dataset_cell_exp[expname]['features'][feature].append(
                            f)

                    dataset_cell_exp[expname]['features']['numspikes'].append(
                        numspike)

                    # Print individual features to table if required
                    if print_table_flag:

                        crr_filename = dataset_cell_exp[
                            expname]['filename'][i_seg]
                        all_feat_filename = os.path.join(
                            self.mainname, 'all_feature_table.txt')
                        if 'num_events' not in self.options['print_table']:
                            multvalnum = 5
                        else:
                            multvalnum = self.options['print_table'][
                                'num_events']
                        tabletools.printFeatures.dump_features(
                            all_feat_filename=all_feat_filename,
                            cellname=cellname,
                            trace_filename=crr_filename,
                            features_name=features_all_,
                            fel_vals=fel_vals,
                            multvalnum=multvalnum,
                            metadata=self.metadataset[
                                cellname]['experiments'][
                                    expname][crr_filename],
                            amp=amp,
                            stim_start=trace['stim_start'][0],
                            stim_end=trace['stim_end'][0],
                        )

    def mean_features(self):
        """Compute the mean for each features for each target"""
        logger.info(" Calculating mean features")

        # mean for each cell
        for i_cell, cellname in enumerate(self.dataset):

            dataset_cell_exp = self.dataset[cellname]['experiments']

            # compute threshold based on some experiments
            if 'expthreshold' in self.options:
                hypamp = []
                amp = []
                numspikes = []

                for i_exp, expname in enumerate(dataset_cell_exp):
                    # use to determine threshold
                    if expname in self.options['expthreshold']:
                        hypamp = hypamp + dataset_cell_exp[expname]['hypamp']
                        amp = amp + dataset_cell_exp[expname]['amp']
                        numspikes = numspikes + dataset_cell_exp[expname][
                            'features']['numspikes']

                mean_hypamp = self.newmeancell(numpy.array(hypamp))
                amp_threshold = self.get_threshold(amp, numspikes)

                self.thresholds_per_cell[cellname] = amp_threshold
                self.hypamps_per_cell[cellname] = mean_hypamp

                logger.info(
                    " %s threshold amplitude: %f hypamp: %f",
                    cellname,
                    amp_threshold,
                    mean_hypamp)

            else:
                self.thresholds_per_cell[cellname] = None
                self.hypamps_per_cell[cellname] = None

            for i_exp, expname in enumerate(dataset_cell_exp):

                # define
                dataset_cell_exp[expname]['mean_amp'] = OrderedDict()
                dataset_cell_exp[expname]['mean_amp_rel'] = OrderedDict()
                dataset_cell_exp[expname]['std_amp'] = OrderedDict()
                dataset_cell_exp[expname]['std_amp_rel'] = OrderedDict()

                dataset_cell_exp[expname]['mean_hypamp'] = OrderedDict()
                dataset_cell_exp[expname]['std_hypamp'] = OrderedDict()

                dataset_cell_exp[expname]['mean_features'] = OrderedDict()
                dataset_cell_exp[expname]['std_features'] = OrderedDict()
                dataset_cell_exp[expname]['n'] = OrderedDict()
                dataset_cell_exp[expname]['raw'] = OrderedDict()

                dataset_cell_exp[expname]['bc_mean_features'] = OrderedDict()
                dataset_cell_exp[expname]['bc_std_features'] = OrderedDict()
                dataset_cell_exp[expname]['bc_shift_features'] = OrderedDict()
                dataset_cell_exp[expname]['bc_ld_features'] = OrderedDict()

                for feature in self.features[expname]:
                    dataset_cell_exp[expname]['mean_features'][feature] = \
                        OrderedDict()
                    dataset_cell_exp[expname]['std_features'][feature] = \
                        OrderedDict()
                    dataset_cell_exp[expname]['bc_mean_features'][feature] = \
                        OrderedDict()
                    dataset_cell_exp[expname]['bc_std_features'][feature] = \
                        OrderedDict()
                    dataset_cell_exp[expname]['bc_shift_features'][feature] = \
                        OrderedDict()
                    dataset_cell_exp[expname]['bc_ld_features'][feature] = \
                        OrderedDict()
                    dataset_cell_exp[expname]['n'][feature] = OrderedDict()
                    dataset_cell_exp[expname]['raw'][feature] = OrderedDict()

                ton = dataset_cell_exp[expname]['ton']
                toff = dataset_cell_exp[expname]['toff']
                tend = dataset_cell_exp[expname]['tend']

                ton = self.newmeancell(numpy.array(ton))
                toff = self.newmeancell(numpy.array(toff))
                tend = self.newmeancell(numpy.array(tend))

                hypamp = dataset_cell_exp[expname]['hypamp']
                amp = dataset_cell_exp[expname]['amp']
                numspikes = dataset_cell_exp[expname]['features']['numspikes']
                feature_array = dataset_cell_exp[expname]['features']

                rawfiles_list = dataset_cell_exp[expname]['rawfiles']

                if self.options["relative"]:
                    amp_threshold = self.thresholds_per_cell[cellname]
                    amp_rel = numpy.array(amp / amp_threshold * 100.)
                else:
                    amp_rel = numpy.array(amp)

                # absolute amplitude not relative to hypamp
                amp_abs = numpy.abs(
                    numpy.array(
                        dataset_cell_exp[expname]['amp']) +
                    numpy.array(
                        dataset_cell_exp[expname]['hypamp']))

                i_noinput = numpy.argmin(amp_abs)

                # save amplitude results
                for ti, target in enumerate(self.options["target"]):

                    if target == 'noinput':
                        idx = numpy.array(i_noinput)
                    elif target == 'all':
                        idx = numpy.ones(len(amp), dtype=bool)
                    else:
                        idx = numpy.array(
                            (amp_rel >= (
                                target -
                                self.options["tolerance"][ti])) & (
                                amp_rel <= (
                                    target +
                                    self.options["tolerance"][ti])))

                    amp_target = numpy.atleast_1d(numpy.array(amp)[idx])
                    # equal to amp_target if amplitude not measured relative to
                    # threshold
                    amp_rel_target = numpy.atleast_1d(
                        numpy.array(amp_rel)[idx])
                    hypamp_target = numpy.atleast_1d(numpy.array(hypamp)[idx])

                    if len(amp_target) > 0:

                        if (target == 'noinput'):
                            if (amp_abs[i_noinput] < 0.01):
                                meanamp_target = self.newmeancell(amp_target)
                                stdamp_target = self.newstdcell(amp_target)
                                meanamp_rel_target = self.newmeancell(
                                    amp_rel_target)
                                stdamp_rel_target = self.newstdcell(
                                    amp_rel_target)
                                meanhypamp_target = self.newmeancell(
                                    hypamp_target)
                                stdhypamp_target = self.newstdcell(
                                    hypamp_target)
                            else:
                                continue

                        else:
                            meanamp_target = self.newmeancell(amp_target)
                            stdamp_target = self.newstdcell(amp_target)
                            meanamp_rel_target = self.newmeancell(
                                amp_rel_target)
                            stdamp_rel_target = self.newstdcell(amp_rel_target)
                            meanhypamp_target = self.newmeancell(hypamp_target)
                            stdhypamp_target = self.newstdcell(hypamp_target)

                        dataset_cell_exp[expname]['mean_amp'][
                            str(target)] = meanamp_target
                        dataset_cell_exp[expname]['std_amp'][
                            str(target)] = stdamp_target

                        dataset_cell_exp[expname]['mean_amp_rel'][
                            str(target)] = meanamp_rel_target
                        dataset_cell_exp[expname]['std_amp_rel'][
                            str(target)] = stdamp_rel_target

                        dataset_cell_exp[expname]['mean_hypamp'][
                            str(target)] = meanhypamp_target
                        dataset_cell_exp[expname]['std_hypamp'][
                            str(target)] = stdhypamp_target

                dataset_cell_exp[expname]['mean_ton'] = ton
                dataset_cell_exp[expname]['mean_toff'] = toff
                dataset_cell_exp[expname]['mean_tend'] = tend

                for fi, feature in enumerate(self.features[expname]):
                    feat_vals = numpy.array(feature_array[feature])

                    for ti, target in enumerate(self.options["target"]):

                        if target == 'noinput':
                            idx = numpy.array(i_noinput)
                        elif target == 'all':
                            idx = numpy.ones(len(amp), dtype=bool)
                        else:
                            idx = numpy.array(
                                (amp_rel >= (
                                    target -
                                    self.options["tolerance"][ti])) & (
                                    amp_rel <= (
                                        target +
                                        self.options["tolerance"][ti])))

                        feat = numpy.atleast_1d(numpy.array(feat_vals)[idx])

                        if self.saveraw:
                            raw = numpy.atleast_1d(
                                numpy.array(rawfiles_list)[idx]).tolist()

                        n = numpy.sum(
                            numpy.invert(
                                numpy.isnan(
                                    numpy.atleast_1d(feat))))

                        if n > 0:

                            if (target == 'noinput'):
                                if (amp_abs[i_noinput] < 0.01):
                                    meanfeat = self.newmeancell(feat)
                                    stdfeat = self.newstdcell(feat)
                                    bcmean, bcstd, bcld, bcshift = \
                                        self.boxcoxcell(
                                            feat, nanopt="nanmean_cell")
                                else:
                                    continue
                            else:
                                meanfeat = self.newmeancell(feat)
                                stdfeat = self.newstdcell(feat)
                                bcmean, bcstd, bcld, bcshift = self.boxcoxcell(
                                    feat, nanopt="nanmean_cell")

                            dataset_cell_exp[expname]['mean_features'][
                                feature][
                                str(target)] = meanfeat
                            dataset_cell_exp[expname]['std_features'][
                                feature][
                                str(target)] = stdfeat
                            dataset_cell_exp[expname]['n'][feature][
                                str(target)] = n

                            if self.saveraw:
                                dataset_cell_exp[expname]['raw'][feature][
                                    str(target)] = raw

                            dataset_cell_exp[expname]['bc_mean_features'][
                                feature][
                                str(target)] = bcmean
                            dataset_cell_exp[expname]['bc_std_features'][
                                feature][
                                str(target)] = bcstd
                            dataset_cell_exp[expname]['bc_shift_features'][
                                feature][
                                str(target)] = bcshift
                            dataset_cell_exp[expname]['bc_ld_features'][
                                feature][
                                str(target)] = bcld

        # mean for all cells
        for i_exp, expname in enumerate(self.experiments):

            # collect everything in global structure
            self.dataset_mean[expname] = OrderedDict()
            self.dataset_mean[expname]['amp'] = OrderedDict()
            self.dataset_mean[expname]['amp_rel'] = OrderedDict()
            self.dataset_mean[expname]['hypamp'] = OrderedDict()
            self.dataset_mean[expname]['ton'] = []
            self.dataset_mean[expname]['toff'] = []
            self.dataset_mean[expname]['tend'] = []
            self.dataset_mean[expname]['features'] = OrderedDict()
            self.dataset_mean[expname]['cell_std_features'] = OrderedDict()
            self.dataset_mean[expname]['cell_bc_features'] = OrderedDict()
            self.dataset_mean[expname]['cell_n'] = OrderedDict()
            self.dataset_mean[expname]['n'] = OrderedDict()
            self.dataset_mean[expname]['raw'] = OrderedDict()

            for feature in self.features[expname]:
                self.dataset_mean[expname]['features'][feature] = OrderedDict()
                self.dataset_mean[expname]['cell_std_features'][feature] = \
                    OrderedDict()
                self.dataset_mean[expname]['cell_bc_features'][feature] = \
                    OrderedDict()
                self.dataset_mean[expname]['cell_n'][feature] = OrderedDict()
                self.dataset_mean[expname]['n'][feature] = OrderedDict()
                self.dataset_mean[expname]['raw'][feature] = OrderedDict()

                for target in self.options["target"]:
                    self.dataset_mean[expname]['features'][feature][
                        str(target)] = []
                    self.dataset_mean[expname][
                        'cell_std_features'][feature][str(target)] = []
                    self.dataset_mean[expname][
                        'cell_bc_features'][feature][str(target)] = []
                    self.dataset_mean[expname]['cell_n'][feature][
                        str(target)] = []
                    self.dataset_mean[expname]['n'][feature][str(target)] = []
                    self.dataset_mean[expname]['raw'][feature][
                        str(target)] = []

            for target in self.options["target"]:
                self.dataset_mean[expname]['amp'][str(target)] = []
                self.dataset_mean[expname]['amp_rel'][str(target)] = []
                self.dataset_mean[expname]['hypamp'][str(target)] = []

            for i_cell, cellname in enumerate(self.dataset):

                dataset_cell_exp = self.dataset[cellname]['experiments']

                if expname in dataset_cell_exp:
                    self.dataset_mean[expname]['location'] =\
                        dataset_cell_exp[expname]['location']

                    ton = dataset_cell_exp[expname]['mean_ton']
                    self.dataset_mean[expname]['ton'].append(ton)

                    toff = dataset_cell_exp[expname]['mean_toff']
                    self.dataset_mean[expname]['toff'].append(toff)

                    tend = dataset_cell_exp[expname]['mean_tend']
                    self.dataset_mean[expname]['tend'].append(tend)

                    for target in self.options["target"]:
                        if str(target) in dataset_cell_exp[expname][
                                'mean_amp']:
                            amp = dataset_cell_exp[expname]['mean_amp'][
                                str(target)]
                            self.dataset_mean[expname]['amp'][
                                str(target)].append(amp)
                            amp_rel = dataset_cell_exp[expname][
                                'mean_amp_rel'][
                                str(target)]
                            self.dataset_mean[expname]['amp_rel'][
                                str(target)].append(amp_rel)
                            hypamp = dataset_cell_exp[expname][
                                'mean_hypamp'][str(target)]
                            self.dataset_mean[expname]['hypamp'][
                                str(target)].append(hypamp)

                    for feature in self.features[expname]:
                        for target in self.options["target"]:
                            if str(target) in dataset_cell_exp[expname][
                                    'mean_features'][feature]:

                                result = dataset_cell_exp[expname][
                                    'mean_features'][feature][
                                    str(target)]
                                self.dataset_mean[expname]['features'][
                                    feature][str(target)].append(result)

                                cell_std_result = dataset_cell_exp[expname][
                                    'std_features'][feature][str(target)]
                                self.dataset_mean[expname][
                                    'cell_std_features'][feature][
                                    str(target)].append(cell_std_result)

                                bcmean = dataset_cell_exp[expname][
                                    'bc_mean_features'][feature][
                                    str(target)]
                                bcstd = dataset_cell_exp[expname][
                                    'bc_std_features'][feature][
                                    str(target)]
                                bcshift = dataset_cell_exp[expname][
                                    'bc_shift_features'][feature][
                                    str(target)]
                                bcld = dataset_cell_exp[expname][
                                    'bc_ld_features'][feature][
                                    str(target)]
                                self.dataset_mean[expname][
                                    'cell_bc_features'][feature][str(
                                        target)].append(
                                    [bcmean, bcstd, bcld, bcshift])

                                n = dataset_cell_exp[expname]['n'][feature][
                                    str(target)]
                                self.dataset_mean[expname][
                                    'cell_n'][feature][
                                    str(target)].append(n)

                                if self.saveraw:
                                    raw = dataset_cell_exp[expname]['raw'][
                                        feature][
                                        str(target)]
                                    self.dataset_mean[expname]['raw'][
                                        feature][
                                        str(target)].append(raw)

            # create means
            self.dataset_mean[expname]['mean_amp'] = OrderedDict()
            self.dataset_mean[expname]['mean_amp_rel'] = OrderedDict()
            self.dataset_mean[expname]['mean_hypamp'] = OrderedDict()
            self.dataset_mean[expname]['std_amp'] = OrderedDict()
            self.dataset_mean[expname]['std_amp_rel'] = OrderedDict()
            self.dataset_mean[expname]['std_hypamp'] = OrderedDict()
            self.dataset_mean[expname]['mean_features'] = OrderedDict()
            self.dataset_mean[expname]['std_features'] = OrderedDict()

            self.dataset_mean[expname]['bc_mean_features'] = OrderedDict()
            self.dataset_mean[expname]['bc_std_features'] = OrderedDict()
            self.dataset_mean[expname]['bc_shift_features'] = OrderedDict()
            self.dataset_mean[expname]['bc_ld_features'] = OrderedDict()

            for feature in self.features[expname]:
                self.dataset_mean[expname][
                    'mean_features'][feature] = OrderedDict()
                self.dataset_mean[expname][
                    'std_features'][feature] = OrderedDict()

                self.dataset_mean[expname][
                    'bc_mean_features'][feature] = OrderedDict()
                self.dataset_mean[expname][
                    'bc_std_features'][feature] = OrderedDict()
                self.dataset_mean[expname][
                    'bc_shift_features'][feature] = OrderedDict()
                self.dataset_mean[expname][
                    'bc_ld_features'][feature] = OrderedDict()

            ton = self.dataset_mean[expname]['ton']
            toff = self.dataset_mean[expname]['toff']
            tend = self.dataset_mean[expname]['tend']

            self.dataset_mean[expname]['mean_ton'] = self.newmean(ton)
            self.dataset_mean[expname]['mean_toff'] = self.newmean(toff)
            self.dataset_mean[expname]['mean_tend'] = self.newmean(tend)

            for target in self.options["target"]:
                amp = self.dataset_mean[expname]['amp'][str(target)]
                amp_rel = self.dataset_mean[expname]['amp_rel'][str(target)]
                hypamp = self.dataset_mean[expname]['hypamp'][str(target)]

                self.dataset_mean[expname]['mean_amp'][
                    str(target)] = self.newmean(amp)
                self.dataset_mean[expname]['mean_amp_rel'][
                    str(target)] = self.newmean(amp_rel)
                self.dataset_mean[expname]['mean_hypamp'][
                    str(target)] = self.newmean(hypamp)
                self.dataset_mean[expname]['std_amp'][
                    str(target)] = self.newstd(amp)
                self.dataset_mean[expname]['std_amp_rel'][
                    str(target)] = self.newstd(amp_rel)
                self.dataset_mean[expname]['std_hypamp'][
                    str(target)] = self.newstd(hypamp)

            for feature in self.features[expname]:
                for target in self.options["target"]:
                    feat = self.dataset_mean[expname]['features'][feature][
                        str(target)]
                    cell_std_feat = self.dataset_mean[expname][
                        'cell_std_features'][feature][str(target)]
                    cell_n = self.dataset_mean[expname]['cell_n'][feature][
                        str(target)]

                    # added by Luca Leonardo Bologna to handle the case in
                    # which only one feature value is present at this point
                    # count nonNan entries!

                    n = numpy.sum(
                        numpy.invert(
                            numpy.isnan(
                                numpy.atleast_1d(feat))))
                    self.dataset_mean[expname]['n'][feature][str(target)] = n

                    if n == 1:  # only result from one cell in population
                        # pick values from this cell
                        val_idx = numpy.where(~numpy.isnan(
                            numpy.atleast_1d(feat)))[0][0]
                        self.dataset_mean[expname]['mean_features'][
                            feature][str(target)] = feat[val_idx]
                        self.dataset_mean[expname]['std_features'][
                            feature][str(target)] = cell_std_feat[val_idx]
                        [bcmean, bcstd, bcld, bcshift] = \
                            self.dataset_mean[expname][
                            'cell_bc_features'][feature][str(
                                target)][val_idx]
                    else:
                        self.dataset_mean[expname][
                            'mean_features'][feature][str(
                                target)] = self.newmean(feat)
                        self.dataset_mean[expname]['std_features'][feature][
                            str(target)] = self.newstd(feat)
                        bcmean, bcstd, bcld, bcshift = self.boxcoxcell(
                            feat, nanopt="nanmean")

                    self.dataset_mean[expname]['bc_mean_features'][feature][
                        str(target)] = bcmean
                    self.dataset_mean[expname]['bc_std_features'][feature][
                        str(target)] = bcstd
                    self.dataset_mean[expname][
                        'bc_shift_features'][feature][str(
                            target)] = bcshift
                    self.dataset_mean[expname]['bc_ld_features'][feature][
                        str(target)] = bcld

    def get_threshold(self, amp, numspikes):
        """Get the spiking threshold of a cell by taking the smallest current
        amplitude for which it fires"""
        isort = numpy.argsort(amp)
        amps_sort = numpy.array(amp)[isort]
        numspikes_sort = numpy.array(numspikes)[isort]
        i_threshold = numpy.where(
            numspikes_sort >= self.options["spike_threshold"])[0][0]
        amp_threshold = amps_sort[i_threshold]

        return amp_threshold

    def plt_features(self):
        """Plot the features"""
        logger.info(" Plotting features")
        figs = OrderedDict()

        markercyclercell = cycle(self.markerlist)
        colorcyclercell = cycle(self.colorlist)

        for i_cell, cellname in enumerate(self.dataset):

            cellfigs = OrderedDict()

            dirname = self.maindirname + cellname
            tools.makedir(dirname)

            colorcell = next(colorcyclercell)
            markercell = next(markercyclercell)
            dataset_cell_exp = self.dataset[cellname]['experiments']

            for i_exp, expname in enumerate(dataset_cell_exp):

                figname = "features_" + expname
                if figname not in figs:
                    plottools.tiled_figure(
                        figname, frames=len(self.features[expname]),
                        columns=3, figs=figs, dirname=self.maindirname,
                        top=0.97, bottom=0.04, left=0.07, right=0.97,
                        hspace=0.75, wspace=0.3)

                figname = "features_" + cellname.split('/')[-1] + "_" + expname
                axs_cell = plottools.tiled_figure(
                    figname,
                    frames=len(
                        self.features[expname]),
                    columns=3,
                    figs=cellfigs,
                    dirname=dirname,
                    top=0.97,
                    bottom=0.04,
                    left=0.07,
                    right=0.97,
                    hspace=0.75,
                    wspace=0.3)

                amp = dataset_cell_exp[expname]['amp']
                feature_array = dataset_cell_exp[expname]['features']
                filenames = dataset_cell_exp[expname]['filename']

                markercycler = cycle(self.markerlist)
                colorcycler = cycle(self.colorlist)

                colormarker_dict = {u: [next(colorcycler), next(
                    markercycler)] for u in list(set(filenames))}

                if self.options["relative"]:
                    amp_threshold = self.thresholds_per_cell[cellname]
                    amp_rel = amp / amp_threshold * 100.
                else:
                    amp_rel = amp

                for fi, feature in enumerate(self.features[expname]):

                    feat_vals = numpy.array(feature_array[feature])

                    for i_feat, feat_vals_ in enumerate(feat_vals):
                        color = colormarker_dict[filenames[i_feat]][0]
                        marker = colormarker_dict[filenames[i_feat]][1]

                        amp_rel_ = numpy.float64(amp_rel[i_feat])
                        is_not_nan = ~numpy.isnan(feat_vals_)

                        axs_cell[fi].plot(
                            amp_rel_[is_not_nan],
                            feat_vals_[is_not_nan],
                            "",
                            linestyle='None',
                            marker=marker,
                            color=color,
                            markersize=5,
                            zorder=1,
                            linewidth=1,
                            markeredgecolor='none',
                            clip_on=False)
                        # axs_cell[fi].set_xticks(self.options["target"])
                        # axs_cell[fi].set_xticklabels(())
                        axs_cell[fi].set_title(feature)

                        figname = "features_" + expname
                        figs[figname]['axs'][fi].plot(
                            amp_rel_[is_not_nan],
                            feat_vals_[is_not_nan],
                            "", linestyle='None', marker=markercell,
                            color=colorcell, markersize=3, zorder=1,
                            linewidth=1, markeredgecolor='none', clip_on=False)
                        figs[figname]['axs'][fi].set_title(feature)

                    if 'mean_features' in dataset_cell_exp[expname]:

                        amp_rel_list = []
                        mean_list = []
                        std_list = []
                        for target in self.options["target"]:
                            if str(target) in dataset_cell_exp[expname][
                                    'mean_features'][feature]:
                                a = dataset_cell_exp[expname][
                                    'mean_amp_rel'][str(
                                        target)]
                                m = dataset_cell_exp[expname][
                                    'mean_features'][feature][str(
                                        target)]
                                s = dataset_cell_exp[expname][
                                    'std_features'][feature][str(
                                        target)]

                                if "zero_std" in self.options and \
                                        self.options["zero_std"]:
                                    rules = [~numpy.isnan(m)]
                                else:
                                    rules = [
                                        ~numpy.isnan(m), (s > 0.0) or
                                        (m == 0.0)]
                                if all(rules):
                                    amp_rel_list.append(a)
                                    mean_list.append(m)
                                    std_list.append(s)

                        mean_array = numpy.array(mean_list)
                        std_array = numpy.array(std_list)
                        amp_rel_array = numpy.array(amp_rel_list)

                        e = axs_cell[fi].errorbar(
                            amp_rel_array, mean_array, yerr=std_array,
                            marker='s', color='k', linewidth=1,
                            linestyle='None', markersize=6, zorder=10,
                            clip_on=False)
                        # axs_cell[fi].set_xticks(self.options["target"])
                        for b in e[1]:
                            b.set_clip_on(False)

            # close single cell figures
            for i_fig, figname in enumerate(cellfigs):
                fig = cellfigs[figname]
                fig['fig'].savefig(
                    fig['dirname'] + '/' + figname + '.pdf', dpi=300)
                plt.close(fig['fig'])

        for i_exp, expname in enumerate(self.experiments):

            if expname in self.dataset_mean:
                for fi, feature in enumerate(self.features[expname]):
                    amp_rel_list = []
                    mean_list = []
                    std_list = []
                    for target in self.options["target"]:
                        if str(target) in self.dataset_mean[expname][
                                'mean_features'][feature]:
                            a = self.dataset_mean[expname]['mean_amp_rel'][
                                str(target)]
                            m = self.dataset_mean[expname]['mean_features'][
                                feature][
                                str(target)]
                            s = self.dataset_mean[expname]['std_features'][
                                feature][
                                str(target)]

                            if "zero_std" in self.options and \
                                    self.options["zero_std"]:
                                rules = [~numpy.isnan(m)]
                            else:
                                rules = [
                                    ~numpy.isnan(m), (s > 0.0) or
                                    (m == 0.0)]
                            if all(rules):
                                amp_rel_list.append(a)
                                mean_list.append(m)
                                std_list.append(s)

                    mean_array = numpy.array(mean_list)
                    std_array = numpy.array(std_list)
                    amp_rel_array = numpy.array(amp_rel_list)

                    figname = "features_" + expname
                    e = figs[figname]['axs'][fi].errorbar(
                        amp_rel_list, mean_array, yerr=std_array, marker='s',
                        color='k', linewidth=1, linestyle='None', markersize=6,
                        zorder=10, clip_on=False)
                    for b in e[1]:
                        b.set_clip_on(False)

        for i_fig, figname in enumerate(figs):
            fig = figs[figname]
            fig['fig'].savefig(
                fig['dirname'] +
                '/' +
                figname +
                '.pdf',
                dpi=300)
            plt.close(fig['fig'])

    def plt_features_dist(self):
        """Plot the distribution of features"""
        logger.info(" Plotting feature distributions")
        figs = OrderedDict()

        dirname = self.maindirname + 'dists'
        tools.makedir(dirname)

        for i_exp, expname in enumerate(self.experiments):

            figpath = dirname + '/features_' + expname + '.pdf'
            pdf_pages = PdfPages(figpath)

            for feature in self.features[expname]:

                figname = "features_" + expname + "_" + feature
                axs = plottools.tiled_figure(
                    figname, frames=2 * len(self.options["target"]),
                    columns=6, figs=figs, dirname=dirname, top=0.92,
                    bottom=0.04, left=0.07, right=0.97, hspace=0.75,
                    wspace=0.3)

                for it, target in enumerate(self.options["target"]):

                    feat = numpy.array(
                        self.dataset_mean[expname]['features'][feature]
                        [str(target)])

                    ax = axs[2 * it]
                    ax_bc = axs[2 * it + 1]
                    ax.set_title(str(target))
                    ax_bc.set_title(str(target) + " boxcox")
                    feat[numpy.isinf(feat)] = float('nan')

                    bcshift = \
                        self.dataset_mean[expname][
                            'bc_shift_features'][feature][str(
                                target)]
                    bcld = self.dataset_mean[expname]['bc_ld_features'][
                        feature][
                        str(target)]

                    # removing nan values before plotting
                    feat = feat[numpy.logical_not(numpy.isnan(feat))]
                    if (all(numpy.isnan(feat)) is False):

                        ax.hist(feat, max(1, int(len(feat) / 2)),
                                histtype='stepfilled', color='b',
                                edgecolor='none')
                        ax_bc.set_title(
                            str(target) +
                            " boxcox\nld:" +
                            str(bcld) +
                            "\nshift:" +
                            str(bcshift))

                fig = figs[figname]
                fig['fig'].suptitle(feature)
                pdf_pages.savefig(fig['fig'])

            pdf_pages.close()

    def feature_config_all(self, version=None):
        self.create_feature_config(self.maindirname,
                                   self.dataset_mean, version=version)

    def feature_config_cells(self, version=None):
        for i_cell, cellname in enumerate(self.dataset):
            dirname = self.maindirname + cellname
            tools.makedir(dirname)
            dataset_cell_exp = self.dataset[cellname]['experiments']
            self.create_feature_config(self.maindirname + cellname + os.sep,
                                       dataset_cell_exp, version=version)

    def analyse_threshold(self):
        """Get the spiking threshold and holding current for  all cells. Save
        them in hypamp_threshold.json"""
        logger.info(
            " Analysing threshold and hypamp and saving files to %s",
            self.maindirname)

        hyp_th = {}
        for cellname in self.thresholds_per_cell:
            hyp_th[cellname] = OrderedDict()
            hyp_th[cellname]['threshold'] = round(
                self.thresholds_per_cell[cellname], 4)
            hyp_th[cellname]['hypamp'] = round(
                self.hypamps_per_cell[cellname], 4)

        thresholds = [d for k, d in self.thresholds_per_cell.items()]
        hypamps = [d for k, d in self.hypamps_per_cell.items()]
        hyp_th['all'] = OrderedDict()
        hyp_th['all']['threshold'] = [
            round(
                numpy.mean(thresholds), 4), round(
                numpy.std(thresholds), 4)]
        hyp_th['all']['hypamp'] = [
            round(
                numpy.mean(hypamps), 4), round(
                numpy.std(hypamps), 4)]

        json.dump(
            hyp_th,
            open(
                self.maindirname +
                "hypamp_threshold.json",
                'w'),
            indent=4,
            cls=tools.NumpyEncoder)

    def create_feature_config(self, directory, dataset, version=None):
        """Save the efeatures and protocols for each protocol/target combo
        in json file"""
        logger.info(" Saving config files to %s", directory)

        if version == 'legacy':

            indent = 8
            stim = OrderedDict()
            feat = OrderedDict()

            for i_exp, expname in enumerate(self.experiments):
                if expname in dataset:
                    location = dataset[expname]['location']

                    for fi, feature in enumerate(self.features[expname]):
                        for it, target in enumerate(self.options["target"]):

                            if str(target) in dataset[expname][
                                    'mean_features'][feature]:

                                t = str(target)
                                stimname = expname + '_' + t

                                m = round(
                                    dataset[expname]['mean_features']
                                    [feature][str(target)],
                                    4)
                                s = round(
                                    dataset[expname]['std_features']
                                    [feature][str(target)],
                                    4)
                                n = int(dataset[expname]['n']
                                        [feature][str(target)])

                                if "zero_std" in self.options and \
                                        self.options["zero_std"]:
                                    rules = [~numpy.isnan(m)]
                                else:
                                    rules = [
                                        ~numpy.isnan(m), (s > 0.0) or
                                        (m == 0.0)]

                                if all(rules):
                                    if s == 0.0:  # prevent divison by 0
                                        s = 1e-3

                                    if stimname not in stim:
                                        stim[stimname] = OrderedDict()

                                    if stimname not in feat:
                                        feat[stimname] = OrderedDict()

                                    if location not in feat[stimname]:
                                        feat[stimname][location] = \
                                            OrderedDict()

                                    feat[stimname][location][feature] = [m, s]

                                    a = round(
                                        dataset[expname]['mean_amp']
                                        [str(target)],
                                        6)
                                    h = round(
                                        dataset[expname]['mean_hypamp']
                                        [str(target)],
                                        6)
                                    ton = dataset[expname]['mean_ton']
                                    toff = dataset[expname]['mean_toff']
                                    tend = dataset[expname]['mean_tend']

                                    if 'stimuli' not in stim[stimname]:

                                        totduration = round(tend)
                                        delay = round(
                                            self.options["delay"] + ton)
                                        duration = round(toff - ton)

                                        stim[stimname]['stimuli'] = [
                                            OrderedDict([
                                                ("delay", delay),
                                                ("amp", a),
                                                ("duration", duration),
                                                ("totduration", totduration),
                                            ]),
                                            OrderedDict([
                                                ("delay", 0.0),
                                                ("amp", h),
                                                ("duration", totduration),
                                                ("totduration", totduration),
                                            ]),
                                        ]

            stim = OrderedDict([(self.mainname, stim)])
            feat = OrderedDict([(self.mainname, feat)])

        else:

            indent = 6
            stim = OrderedDict()
            feat = OrderedDict()
            featraw = OrderedDict()

            boxcox = False
            if ('boxcox' in self.options) and self.options['boxcox']:
                boxcox = True

            fid = 0
            for i_exp, expname in enumerate(self.experiments):
                if expname in dataset:
                    location = dataset[expname]['location']

                    for fi, feature in enumerate(self.features[expname]):
                        for it, target in enumerate(self.options["target"]):

                            if str(target) in dataset[expname][
                                    'mean_features'][feature]:

                                t = str(target)
                                stimname = expname + '_' + t

                                m = round(
                                    dataset[expname]['mean_features']
                                    [feature][str(target)],
                                    4)
                                s = round(
                                    dataset[expname]['std_features']
                                    [feature][str(target)],
                                    4)
                                n = int(dataset[expname]['n']
                                        [feature][str(target)])

                                if self.saveraw:
                                    raw = dataset[expname]['raw'][feature][
                                        str(target)]

                                bcm = dataset[expname]['bc_mean_features'][
                                    feature][
                                    str(target)]
                                bcs = dataset[expname]['bc_std_features'][
                                    feature][
                                    str(target)]
                                bcshift = dataset[expname][
                                    'bc_shift_features'][feature][
                                    str(target)]
                                bcld = dataset[expname]['bc_ld_features'][
                                    feature][
                                    str(target)]

                                if boxcox:
                                    do_add = ~numpy.isnan(bcm)
                                else:
                                    do_add = ~numpy.isnan(m)

                                if do_add:

                                    if s == 0.0:  # prevent divison by 0
                                        s = 1e-3

                                    if bcs == 0.0:  # prevent divison by 0
                                        bcs = 1e-3

                                    if stimname not in stim:
                                        stim[stimname] = OrderedDict()

                                    if stimname not in feat:
                                        feat[stimname] = OrderedDict()
                                        featraw[stimname] = OrderedDict()

                                    if location not in feat[stimname]:
                                        feat[stimname][location] = []
                                        featraw[stimname][location] = []

                                    if boxcox and (bcshift is not False):
                                        feat[stimname][location].append(
                                            OrderedDict([
                                                ("feature", feature),
                                                ("val",
                                                 [m, s, bcm, bcs,
                                                  bcld, bcshift]),
                                                ("n", n),
                                                ("fid", fid)
                                            ]))

                                        if self.saveraw:
                                            featraw[stimname][location].append(
                                                OrderedDict([
                                                    ("feature", feature),
                                                    ("val", [m, s, bcm, bcs,
                                                             bcld, bcshift]),
                                                    ("n", n),
                                                    ("fid", fid),
                                                    ("raw", raw)
                                                ]))

                                    else:

                                        feat[stimname][location].append(
                                            OrderedDict([
                                                ("feature", feature),
                                                ("val", [m, s]),
                                                ("n", n),
                                                ("fid", fid)
                                            ]))

                                        if self.saveraw:
                                            featraw[stimname][location].append(
                                                OrderedDict([
                                                    ("feature", feature),
                                                    ("val", [m, s]),
                                                    ("n", n),
                                                    ("fid", fid),
                                                    ("raw", raw)
                                                ]))

                                    fid += 1

                                    if expname in self.options[
                                            "strict_stiminterval"].keys(
                                    ):
                                        strict_stiminterval = self.options[
                                            "strict_stiminterval"][expname]
                                    else:
                                        strict_stiminterval = self.options[
                                            "strict_stiminterval"]['base']
                                    feat[stimname][location][-1][
                                        "strict_stim"] = strict_stiminterval

                                    if self.saveraw:
                                        featraw[stimname][location][-1][
                                            "strict_stim"] = \
                                            strict_stiminterval

                                    a = round(
                                        dataset[expname]['mean_amp']
                                        [str(target)],
                                        6)
                                    h = round(
                                        dataset[expname]['mean_hypamp']
                                        [str(target)],
                                        6)
                                    threshold = round(
                                        dataset[expname]
                                        ['mean_amp_rel']
                                        [str(target)],
                                        4)
                                    ton = dataset[expname]['mean_ton']
                                    toff = dataset[expname]['mean_toff']
                                    tend = dataset[expname]['mean_tend']

                                    if 'stimuli' not in stim[stimname]:

                                        try:
                                            totduration = round(tend)
                                        except ValueError:
                                            totduration = numpy.NaN
                                        delay = round(
                                            self.options["delay"] + ton)
                                        duration = round(toff - ton)

                                        # special frequency pulse stimulus
                                        if expname == 'H40S8':
                                            n = 8
                                            duration = 2.5
                                            stim[stimname]['type'] = \
                                                'StepProtocol'
                                            stim[stimname]['stimuli'] = \
                                                OrderedDict()
                                            stim[stimname][
                                                'stimuli']['step'] = []
                                            totduration = delay + n * 25.
                                            # threshold not estimated,
                                            # used fixed values
                                            threshold = 600.

                                            for s in range(n):
                                                stim[stimname]['stimuli'][
                                                    'step'].append(
                                                    OrderedDict(
                                                        [("delay", delay + s *
                                                          25.),
                                                         ("amp", a),
                                                         ("thresh_perc",
                                                          threshold),
                                                         ("duration",
                                                          duration),
                                                         ("totduration",
                                                          totduration), ]))

                                            stim[stimname]['stimuli'][
                                                'holding'] = OrderedDict(
                                                [("delay", 0.0),
                                                 ("amp", h),
                                                 ("duration", totduration),
                                                 ("totduration",
                                                  totduration), ])
                                        else:
                                            stim[stimname]['type'] = \
                                                'StepProtocol'
                                            stim[stimname][
                                                'stimuli'] = OrderedDict([
                                                    ('step',
                                                     OrderedDict([
                                                         ("delay", delay),
                                                         ("amp", a),
                                                         ("thresh_perc",
                                                             threshold),
                                                         ("duration",
                                                             duration),
                                                         ("totduration",
                                                             totduration),
                                                     ])),
                                                    ('holding',
                                                     OrderedDict([
                                                         ("delay", 0.0),
                                                         ("amp", h),
                                                         ("duration",
                                                             totduration),
                                                         ("totduration",
                                                             totduration),
                                                     ])),
                                                ])

        s = json.dumps(stim, indent=2, cls=tools.NumpyEncoder)
        s = tools.collapse_json(s, indent=indent)
        with open(directory + "protocols.json", "w") as f:
            f.write(s)

        s = json.dumps(feat, indent=2, cls=tools.NumpyEncoder)
        s = tools.collapse_json(s, indent=indent)
        with open(directory + "features.json", "w") as f:
            f.write(s)

        if self.saveraw:
            s = json.dumps(featraw, indent=2, cls=tools.NumpyEncoder)
            s = tools.collapse_json(s, indent=indent)
            with gzip.open(directory + "features_sources.json.gz", "wb") as f:
                f.write(s.encode('utf-8'))

    def create_metadataset(self):
        """
        Fill a dictionary with metadata for every file to be processed
        If no metadata file is present, default values are inserted
        """

        logger.info("Filling metadataset")
        path = self.path

        self.metadataset = OrderedDict()

        for i_cell, cellname in enumerate(self.cells):

            self.metadataset[cellname] = OrderedDict()
            self.metadataset[cellname]['path'] = path

            metadataset_cell_exp = OrderedDict()
            self.metadataset[cellname]['experiments'] = metadataset_cell_exp

            for i_exp, expname in enumerate(self.cells[
                    cellname]['experiments']):

                files = self.cells[cellname]['experiments'][expname]['files']

                if len(files) > 0:
                    logger.debug(" Adding experiment %s to metadata", expname)

                    metadataset_cell_exp[expname] = OrderedDict()
                    if self.format == "igor":
                        for dict_igor in files:
                            cellname = dict_igor["ordinal"]
                            foldpath = os.path.dirname(dict_igor["v_file"])
                            fullpath = os.path.join(
                                foldpath, cellname + '_' + 'metadata.json')
                            print(fullpath)
                            metadataset_cell_exp[expname][cellname] = \
                                common.manageMetadata.get_metadata(
                                fullpath)
                    else:
                        for idx_file, filename in enumerate(files):
                            fullpath = os.path.join(
                                self.path, cellname, filename + '_' +
                                'metadata.json')
                            metadataset_cell_exp[expname][filename] = \
                                common.manageMetadata.get_metadata(
                                fullpath)
        return True



"""Axon reader"""

"""
Copyright (c) 2020, EPFL/Blue Brain Project

 This file is part of BluePyEfe <https://github.com/BlueBrain/BluePyEfe>

 This library is free software; you can redistribute it and/or modify it under
 the terms of the GNU Lesser General Public License version 3.0 as published
 by the Free Software Foundation.

 This library is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 details.

 You should have received a copy of the GNU Lesser General Public License
 along with this library; if not, write to the Free Software Foundation, Inc.,
 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
"""

from neo import io
from collections import OrderedDict
import logging
logger = logging.getLogger(__name__)
import numpy as np
import os
import json
import sys
import quantities
from quantities import ms, s, mV, nA
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from neo import AnalogSignal
from neo.io import Spike2IO
from . import common


def process(config=None,
            filename=None,
            cellname=None,
            expname=None,
            stim_feats=None,
            idx_file=None,
            ljp=0, v_corr=0):

    path = config['path']
    cells = config['cells']

    # initialize data dictionary
    data = common.manageDicts.initialize_data_dict()

    # read metadata file if present
    meta_dict = {}

    f_meta = os.path.join(path, cellname, filename + '_metadata.json')

    #
    if os.path.exists(f_meta):
        with open(f_meta) as json_metadata:
            meta_dict = json.load(json_metadata)
        json_metadata.close()
    else:
        logger.debug(
            "No metadata file found for file %s. Skipping file", filename)
        return data

    # read stimulus features if present
    stim_feats = []

    logger.debug(" Adding spike2 file %s", filename)

    fullfilename = filename + '.smr'

    logger.debug("Reading file :" + fullfilename)

    f = os.path.join(path, cellname, fullfilename)

    # read .smr data
    signals = read_data(f)

    # extract parameters from metadata
    gain, voltage_unit, vm_channel, stimulus_start, stimulus_end, \
        stimulus_time_unit, stim_channel, holding_voltage, stimulus_unit, \
        stimulus_threshold = extract_metadata(meta_dict)

    # check parameters
    if None in (gain, voltage_unit, vm_channel, stimulus_start, stimulus_end,
                stimulus_time_unit, stim_channel, stimulus_unit,
                stimulus_threshold):
        logger.debug(
            "Parameters are not valid or not present. Skipping file: " +
            fullfilename
        )
        return data

    # read vm signal
    vm = signals[vm_channel[0]]
    conv_factor = common.manageConfig.conversion_factor("mV", voltage_unit)
    vm = set_units(vm, gain * conv_factor, mV)  # add dimension to signal
    vm = vm.flatten()[::vm_channel[1] + 1]

    # read stim signal
    stim = signals[stim_channel[0]]
    conv_factor = common.manageConfig.conversion_factor("nA", stimulus_unit)
    holding_current = common.manageMetadata.get_holding_current(
        meta_dict, "nA")
    stim = set_units(stim, conv_factor, "nA")  # add dimension to signal

    stim = stim - holding_current["holdcurr"]["value"] * stim.units

    stim = stim.flatten()[::stim_channel[1] + 1]
    stimulus_threshold = stimulus_threshold * conv_factor

    # find edges of the signal where different stimulus steps are located
    step_edges = []
    for idx_sts in range(len(stimulus_start)):
        sts = stimulus_start[idx_sts]
        ste = stimulus_end[idx_sts]
        se = find_stimulus_steps(
            stim.time_slice(sts, ste),
            stimulus_threshold
        )
        if idx_sts == 0:
            step_edges = se
        else:
            se_mag = se.magnitude
            se_unit = se.units
            step_edges_mag = step_edges.magnitude
            mag = np.append(step_edges_mag, se_mag)
            step_edges = mag * se_unit

    if "remove_transient_size" in cells[cellname] and \
            cells[cellname]['remove_transient_size']:
        # remove transient
        vm = remove_transients(
            vm, step_edges, transient_length=cells[cellname]
            ['remove_transient_size'][0][0] * ms
        )

    # extract onsets and offsets from edges
    step_onsets = step_edges[::2]
    step_offsets = step_edges[1::2]

    all_vm = vm.flatten()
    all_stim = stim

    # adding margins to recording chunks
    margins = 50 * ms

    # extract values for individual stimuli
    for onset, offset in zip(step_onsets, step_offsets):

        # slicing current sweep voltage signal
        voltage = all_vm.time_slice(onset - margins, offset + margins)
        voltage = voltage.flatten()

        # slicing current sweep stimulus signal
        stimulus = all_stim.time_slice(
            onset - margins,
            offset + margins
        )

        # extracting stimulus chunks
        stim_no_borders = all_stim.time_slice(
            onset + 0.0 * ms,
            offset - 0.0 * ms
        )

        # extracting time step
        dt = 1. / int(all_vm.sampling_rate) * 1e3

        # time
        t = (voltage.times - voltage.t_start).rescale("ms")

        # extracting onset and offset index from arrays
        ion = time_index(stimulus, onset)
        ioff = time_index(stimulus, offset)

        # convert indices to timestamps
        ton = ion * dt
        toff = ioff * dt

        # duration in ms of signal slice from which to extract the
        # stimulus amplitude
        cut = 5
        amp = extract_amp_from_sig(stim_no_borders, onset, offset, cut)
        amp = np.float64(amp)

        # convert stimulus to string
        amp = np.float64(str("{0:.2f}".format(amp)))

        # creating current signal
        current = []
        current = np.zeros(len(voltage))
        current[ion:ioff] = amp

        # estimate hyperpolarization current
        hypamp = np.mean(current[0:ion])

        # convert voltage from AnalogSignal to list
        voltage = np.array(voltage.tolist()).astype(np.float64)

        # normalize membrane potential to known value
        # (given in UCL excel sheet)
        if isinstance(v_corr, list):
            if len(v_corr) == 1 and v_corr[0] != 0.0:
                voltage = voltage - numpy.mean(voltage[0:ion]) + v_corr[0]
            elif len(v_corr) - 1 >= idx_file and v_corr[0] != 0.0:
                voltage = voltage - numpy.mean(voltage[0:ion]) \
                    + v_corr[idx_file]
        elif v_corr != 0.0:
            voltage = voltage - numpy.mean(voltage[0:ion]) + v_corr

        voltage = voltage - ljp

        # extract stim traces to be excluded
        [crr_exc, crr_exc_u] = common.manageConfig.get_exclude_values(
            cells[cellname], idx_file)

        t = t.magnitude.flatten()

        if not len(crr_exc) == 0 and any(abs(crr_exc - amp) < 1e-4):
            continue  # llb
        else:
            common.manageDicts.fill_dict_single_trace(
                data=data, voltage=voltage, current=current, dt=dt, t=t,
                ton=ton, toff=toff, amp=amp, hypamp=hypamp,
                filename=filename)
    return data


def extract_metadata(metadata):
    """
    Extract metadata needed to read .smr traces
    """

    try:
        gain = float(metadata["gain"])
    except Exception as ex:
        logger.error(
            "Error reading 'gain' in metadata or key not present. " +
            "Storing value: None"
        )
        logger.error(ex)
        gain = None

    try:
        stimulus_threshold = float(metadata["stimulus_threshold"][0])
    except Exception as ex:
        logger.error(
            "Error reading 'stimulus_threshold' in metadata or key not " +
            "present. Storing value: None"
        )
        logger.error(ex)
        stimulus_threshold = None

    try:
        holding_voltage = float(metadata["holding_voltage"][0])
    except Exception as ex:
        logger.error(
            "Error reading 'holding_voltage' in metadata or key not " +
            "present. Storing value: '0.0'"
        )
        logger.error(ex)
        holding_voltage = 0.0

    try:
        voltage_unit = metadata["voltage_unit"]
    except Exception as ex:
        logger.error(
            "Error reading 'voltage_unit' in metadata or key not present. " +
            "Storing value: 'mV'"
        )
        logger.error(ex)
        voltage_unit = "mV"

    try:
        stimulus_time_unit = getattr(
            quantities, metadata["stimulus_time_unit"])
    except Exception as ex:
        logger.error(
            "Error reading 'stimulus_time_unit' in metadata or key not " +
            "present. Storing value: 'None'"
        )
        logger.error(ex)
        stimulus_time_unit = None

    try:
        vm_channel = metadata.get("vm_channel", None)
    except Exception as ex:
        logger.error(
            "Error reading 'vm_channel' in metadata or key not present. " +
            "Storing value: 'None'"
        )
        logger.error(ex)
        vm_channel = None

    try:
        if "stimulus_start" not in metadata.keys():
            raise Exception
        else:
            stim_start = metadata["stimulus_start"]
            if type(stim_start) in (float, int):
                stimulus_start = float(stim_start)
            elif type(stim_start) == list and len(stim_start) >= 1:
                stimulus_start = []
                for i in stim_start:
                    stimulus_start.append(i * stimulus_time_unit)
            else:
                raise Exception
    except Exception as ex:
        logger.error(
            "Error reading 'stimulus_start' in metadata, key not present " +
            "or wrong. Storing value: 'None'"
        )
        logger.error(ex)
        stimulus_start = None

    try:
        if "stimulus_end" not in metadata.keys():
            raise Exception
        else:
            stim_end = metadata["stimulus_end"]
            if type(stim_end) in (float, int):
                stimulus_end = float(stim_end)
            elif type(stim_end) == list and len(stim_end) >= 1:
                stimulus_end = []
                for i in stim_end:
                    stimulus_end.append(i * stimulus_time_unit)
            else:
                raise Exception
    except Exception as ex:
        logger.error(
            "Error reading 'stimulus_end' in metadata, key not present or " +
            "wrong. Storing value: 'None'"
        )
        logger.error(ex)
        stimulus_end = None

    try:
        stim_channel = metadata["stim_channel"]
    except Exception as ex:
        logger.error(
            "Error reading 'stim_channel' in metadata or key not present. " +
            "Storing value: None"
        )
        logger.error(ex)
        stim_channel = None

    try:
        stimulus_unit = metadata["stimulus_unit"]
    except Exception as ex:
        logger.error(
            "Error reading 'stimulus_unit' in metadata or key not present." +
            " Storing value: 'nA'"
        )
        logger.error(ex)
        stimulus_unit = None

    return \
        gain, voltage_unit, vm_channel, stimulus_start, stimulus_end, \
        stimulus_time_unit, stim_channel, holding_voltage, stimulus_unit, \
        stimulus_threshold


def set_units(signal, gain, unit):
    """
    Author: Andrew Davison
    """
    # cannot change the units of a signal once created, so we create a new
    # signal from the original data (original signal is dimensionless)
    orig = signal
    sig = AnalogSignal(
        orig.magnitude * gain,
        units=unit,
        sampling_rate=orig.sampling_rate
    )
    sig._copy_data_complement(orig)

    return sig


def find_stimulus_steps(stimulus, stimulus_threshold):
    """
    Authors: Andrew Davison, Luca L. Bologna
    """
    # cut signals into segments so as to overlay responses to different
    # stimulus steps;
    x = np.abs(stimulus.magnitude.flatten()) > stimulus_threshold
    step_edge_indices = np.nonzero(x[1:] ^ x[:-1])[0]

    # does not include step of amplitude zero.
    step_edges = stimulus.times[step_edge_indices]

    return step_edges


def remove_transients(
        vm_trace, stimulus_step_edges, transient_length=1.0 * ms,
        window=5.0 * ms):

    """
    Replace the voltage transients at stimulus onset and offset by the mean of
    the previous few milliseconds.

    This is a rather crude method, and is problematic for offsets if there is
    an action potential just before offset.
    A better approach might be to fit an exponential to the transient.
    Authors: Andrew Davison, Luca L. Bologna
    """
    vm = AnalogSignal(
        vm_trace.magnitude,
        sampling_rate=vm_trace.sampling_rate,
        units=vm_trace.units
    )
    for t_step in stimulus_step_edges:
        length = int(transient_length / vm.sampling_period.rescale('ms'))
        level = vm.time_slice(t_step - window, t_step).mean()
        splice = AnalogSignal(level * np.ones((length,)),
                              sampling_rate=vm.sampling_rate,
                              units=vm.units,
                              t_start=t_step)
        vm.splice(splice)
    return vm


def read_data(filename):
    """
    Author: Andrew Davison
    """
    io = Spike2IO(filename)
    data = io.read()

    return data[0].segments[0].analogsignals


def extract_amp_from_sig(signal, onset, offset, cut):

    """
    Extract amplitude value from signal
    """

    signal_cut = signal.time_slice(onset, onset + cut * ms)
    signal_cut = signal_cut.flatten().magnitude.tolist()
    amp = float(max(signal_cut, key=signal_cut.count))

    return amp


# The following function has been taken from the neo-python package
# Copyright (c) 2010-2018, Neo authors and contributors
# All rights reserved.
def time_index(signal, t):
    """Return the array index corresponding to the time `t`"""
    t = t.rescale(signal.sampling_period.units)
    i = (t - signal.t_start) / signal.sampling_period
    i = int(np.rint(i.magnitude))

    return i
